import os
import sys
import numpy as np
from mayavi.core.ui.api import MayaviScene, SceneEditor, MlabSceneModel
from traits.api import (Bool, Button, cached_property, File, HasTraits,
    Instance, on_trait_change, Str, Property, Directory, Dict, DelegatesTo,
    HasPrivateTraits, Any, List, Enum, Int, Event, Float, Tuple, Range,
    Color)
from traitsui.api import (View, Item, Group, OKCancelButtons, ShellEditor,
    HGroup, VGroup, InstanceEditor, TextEditor, ListEditor, CSVListEditor,
    Handler, Label, OKCancelButtons, VSplit)
from traitsui.message import error as error_dialog
from traitsui.api import MenuBar, Menu, Action

from .custom_list_editor import CustomListEditor

from .electrode import Electrode
from .name_holder import NameHolder, GeometryNameHolder, NameHolderDisplayer
from .utils import (virtual_points3d, crash_if_freesurfer_is_not_sourced,
    gensym, get_subjects_dir, intize)
from .color_utils import mayavi2traits_color
from .geometry import load_affine
from functools import partial
import nibabel as nib



class ElectrodePositionsModel(HasPrivateTraits):
#class ElectrodePositionsModel(HasTraits):
    ct_scan = File
    t1_scan = File
    subjects_dir = Directory
    def _subjects_dir_default(self):
        if 'SUBJECTS_DIR' in os.environ:
            return os.environ['SUBJECTS_DIR']
        else:
            return ''
    subject = Str
    def _subject_default(self):
        if 'SUBJECT' in os.environ:
            return os.environ['SUBJECT']
        else:
            return ''
    fsdir_writable = Bool

    use_ct_mask = Bool(False)
    disable_erosion = Bool(False)
    overwrite_xfms = Bool(False)
    registration_procedure = Enum('uncorrected MI registration',
        'experimental shape correction', 'no registration')
    registration_algorithm = Enum('normalized mutual information', 
        'mutual information', 'rigid')
    shapereg_slice_diff = Float(5.0)

    #this was previously marked as transient which did not cause any problem
    #is there a reason it needed to be transient?
    zoom_factor_override = List(Float, [1.0, 1.0, 1.4] )

    electrode_geometry = List(List(Int), [[8,8]]) # Gx2 list

    electrodes = List(Electrode)
    #interactive_mode = Instance(NameHolder)
    interactive_mode = Property
    def _get_interactive_mode(self):
        return self.interactive_mode_displayer.interactive_mode
    #def _set_interactive_mode(self, val):
    #    self.interactive_mode_displayer.interactive_mode = val

    interactive_mode_displayer = Instance(NameHolderDisplayer, (),
        transient=False)

    _grids = Dict # Str -> List(Electrode)
    #_grid_named_objects = Property(depends_on='_grids')
    #_grid_named_objects = List(NameHolder)
    
    #@on_trait_change('_grid_named_objects')
    #def _update_interactive_mode_displayer(self):
    #    self.interactive_mode_displayer.name_holders = (
    #        self._grid_named_objects)
    #def _update_grid_named_objects_force(self):
    #    swap = self._grid_named_objects
    #    self._grid_named_objects = []
    #    self._grid_named_objects = swap

    surface_opacity = Float(0.4)
        
    _sorted_electrodes = Dict() # Tuple -> Electrode
    _interpolated_electrodes = Dict() # Tuple -> Electrode
    _unsorted_electrodes = Dict() # Tuple -> Electrode
    _all_electrodes = Dict() # Tuple -> Electrode
        # dictionary from surface coordinates (as hashable) to reused
        # electrode objects

    _surf_to_iso_map = Dict()
    _iso_to_surf_map = Dict()

    _iso_to_grid_ident_map = Dict() # Tuple -> Str
    
    _points_to_cur_grid = Dict()
    _points_to_unsorted = Dict()

    _single_glyph_to_recolor = Tuple()
    _new_glyph_color = Any()
    _new_ras_positions = Dict()

    _interactive_mode_snapshot = Str()

    _rebuild_vizpanel_event = Event
    _rebuild_guipanel_event = Event
    _update_glyph_lut_event = Event
    _update_single_glyph_event = Event
    _reorient_glyph_event = Event

    _visualization_ready = Bool(False, )

    _add_annotation_event = Event
    _add_label_event = Event
    _remove_labels_event = Event 
    _label_file = Str()
    _label_borders = Bool(True, )
    _label_opacity = Range(0., 1., 1., )
    _label_color = Color()
    _label_hemi = Enum('both','lh','rh', )

    _hide_noise_event = Event
    _noise_hidden = Bool(False, )

    _colors = Any() # OrderedDict(Grid -> Color)
    _color_scheme = Any(transient=True) #Generator returning 3-tuples
    _grid_geom = Dict() # Grid -> Gx2 list
    _grid_types = Dict() # Grid -> Str

    ct_registration = File

    ct_threshold = Float(2500.)
    dilation_iterations = Int(25)

    critical_percentage = Range(0., 1., 0.75)

    delta = Float(0.5)
    epsilon = Float(10.)
    rho = Float(35.)
    rho_strict = Float(20.)
    rho_loose = Float(50.)

    delta_recon = Float(0.65)
    epsilon_recon = Float(10.)
    rho_recon = Float(40.)
    rho_strict_recon = Float(30.)
    rho_loose_recon = Float(55.)

    isotropize = Enum('By header', 'By voxel', 'Manual override',
        'Isotropization off')

    #this was previously marked as transient which caused a problem with
    #adding new points after load. is there a reason it needs to be transient?
    isotropization_override = List(Float, [1.0, 1.0, 2.5] )

    roi_parcellation = Str('aparc')
    roi_error_radius = Float(4.)

    coronal_dpi = Float(125.)
    coronal_size = List(Float)
    def _coronal_size_default(self):
        return [450., 450.]

    _snapping_completed = Bool(False, )
    deformation_constant = Float(1.)
    sa_steps_break = Int(2500)
    sa_steps_total = Int(2500)
    sa_init_temp = Float(1e-3)
    sa_exp = Float(1.)

    #state-storing interactive labeling windows
    ews = Dict(transient=True) #str -> Instance(HasTraits)
    alw = Instance(HasTraits, transient=True)
    #raw = Instance(HasTraits)
    
    panel2d = Instance(HasTraits, transient=True)
    _cursor_tracker = Instance(Electrode, transient=True)
    

    #do special operations on old pickle types
#    def __setstate__(self, d):
#        self.__dict__ = d
#        if type(d['zoom_factor_override'] == float:
#            self.zoom_factor_override = [1.0, 1.0, d['zoom_factor_override']]

    def __setstate__(self, pickle_dict):
        self._init_trait_listeners()

        for field in pickle_dict: 
            try:
                #setattr(self, field, getattr(pickle_dict, field)) 
                setattr(self, field, pickle_dict[field])
            except Exception as e:
                print(('Failed to load field {0} from pickle file '
                    ''.format(field)))
                print(e)
        
        #print self.__dict__

    def _create_default_name_holders(self):
    #    return self._get__grid_named_objects()
        name_holders = [NameHolder(name=''), 
            GeometryNameHolder(name='unsorted',
                geometry='n/a',
                #TODO dont totally hardcode this color
                color=mayavi2traits_color((1,0,0)))]

        # put additional grids in
        if self._colors is not None:
            for key in list(self._colors.keys()):
                if key in ('unsorted','selection'):
                    continue
                name_holders.append( self._new_grid_name_holder(key) )

        return name_holders

    def _rebuild_interactive_mode_displayer(self, previous_holder=None):
        if previous_holder is not None:
            holder_index = self.interactive_mode_displayer.name_holders.index(
                previous_holder)

        self.interactive_mode_displayer = NameHolderDisplayer()

        self.interactive_mode_displayer.name_holders = (
            self._create_default_name_holders())

        #self.interactive_mode_displayer.interactive_mode = (
        #    self.interactive_mode)

        if previous_holder is not None:
            self.interactive_mode_displayer.interactive_mode = (
                self.interactive_mode_displayer.name_holders[holder_index])

        self._rebuild_guipanel_event = True

    #grid named objects is broken, does not update on reload and does not
    #fully update on grid adding
    #@cached_property
#    def _get__grid_named_objects(self):
#        from color_utils import mayavi2traits_color
#        grid_names = [NameHolder(name=''), 
#            GeometryNameHolder(name='unsorted',
#                geometry='n/a',
#                #TODO dont totally hardcode this color
#                color=mayavi2traits_color((1,0,0)))]
#
#        #for key in sorted(self._grids.keys()):
#        #use the canonical order as the order to appear in the list
#        if self._colors is not None:
#            for key in self._colors.keys():
#                if key in ('unsorted','selection'):
#                    continue
#                grid_names.append(GeometryNameHolder(
#                    name=key, 
#                    geometry=str(self._grid_geom[key]), 
#                    color=mayavi2traits_color(self._colors[key])))
#
#        #if len(self._grids) > 0:
#        #import pdb
#        #pdb.set_trace()
#
#        return grid_names

    def _new_grid_name_holder(self, key):
        gnh = GeometryNameHolder(
            name=key,
            previous_name=key,
            geometry=str(self._grid_geom[key]),
            color=mayavi2traits_color(self._colors[key])) 

        #gnh.on_trait_change(partial(self._change_grid_name, gnh),name='name')
        gnh.on_trait_change( lambda:self._change_grid_name(gnh), name='name')

        return gnh

    def _change_grid_name(self, holder):
        self._commit_grid_changes()

        import copy
        old_name = holder.previous_name
        new_name = holder.name

        if old_name == new_name:
            return

        if new_name in self._grids:
            error_dialog('That name already exists. Pick a different one')  
            return

        if new_name == '':
            error_dialog('Please specify a non-empty name')

        self._grids[new_name] = copy.copy(self._grids[old_name])
        del self._grids[old_name]

        #make sure colors as ordereddict stays in correct order
        #otherwise visualizations will get confused
        from collections import OrderedDict

        colors_index = list(self._colors.keys()).index(old_name)
        new_colors_dict = OrderedDict()
        for i, key in enumerate(self._colors):
            if i == colors_index:
                new_colors_dict[new_name] = self._colors[key]
            else:
                new_colors_dict[key] = self._colors[key]
        self._colors = new_colors_dict 

        self._grid_geom[new_name] = copy.copy(self._grid_geom[old_name])
        del self._grid_geom[old_name]

        self._grid_types[new_name] = copy.copy(self._grid_types[old_name])
        del self._grid_types[old_name]

        for electrode in self._grids[new_name]:
            electrode.grid_name = new_name

        holder.previous_name = new_name

        #self._update_glyph_lut_event = True

        self._rebuild_interactive_mode_displayer(previous_holder=holder)

    def _interactive_mode_changed(self):
        self._commit_grid_changes()

    @on_trait_change('interactive_mode_displayer:_mode_changed_event')
    def _commit_grid_changes(self):
        for p in (self._points_to_cur_grid, self._points_to_unsorted):
            for loc in p:
                elec = p[loc]
                
                old = elec.grid_name
                new = elec.grid_transition_to

                elec.grid_name = new
                elec.grid_transition_to = ''
        
                if old not in ('','unsorted','selection'):
                    self._grids[old].remove(elec)
                    self._iso_to_grid_ident_map[ intize(elec.asiso()) ] = ( 
                        'unsorted')
                elif old=='unsorted':
                    del self._unsorted_electrodes[ intize(elec.asiso()) ]

                if new not in ('','unsorted','selection'):
                    self._grids[new].append(elec)
                    self._iso_to_grid_ident_map[ intize(elec.asiso()) ] = new
                elif new=='unsorted':
                    self._unsorted_electrodes[ intize(elec.asiso()) ] = elec

        self._points_to_cur_grid = {}
        self._points_to_unsorted = {}
    
    def acquire_affine(self):
        from . import pipeline as pipe

        overwrite = self.overwrite_xfms and not self._visualization_ready

        registration_strings = {'mutual information' : ['--cost', 'mi'],
                    'normalized mutual information' : ['--cost', 'nmi'], 
                    'rigid' : ['']}

        if self.ct_registration not in (None, ''):
            aff = load_affine(self.ct_registration)

        elif self.registration_procedure == 'experimental shape correction':
            aff = pipe.register_ct_using_zoom_correction(
                self.ct_scan, subjects_dir=self.subjects_dir,
                subject=self.subject, overwrite=overwrite,
                zf_override=self.zoom_factor_override,
                registration_algorithm=registration_strings[self.registration_algorithm])

        elif self.registration_procedure == 'uncorrected MI registration':
            aff = pipe.register_ct_to_mr_using_mutual_information(
                self.ct_scan, subjects_dir=self.subjects_dir, 
                subject=self.subject, overwrite=overwrite,
                registration_algorithm=registration_strings[self.registration_algorithm])

        #recommend that the user never do this
        elif self.registration_procedure == 'no registration':
            #aff = np.eye(4)

            aff = pipe.get_rawavg_to_orig_xfm(
                subjects_dir=self.subjects_dir,
                subject=self.subject)

            #aff = np.array(((-1., 0., 0., 128.),
            #                (0., 0., 1., -128.),
            #                (0., -1., 0., 128.),
            #                (0., 0., 0., 1.)))

            #aff = np.array(((1., 0., 0., 0.),
            #                (0., 0., -1., 0.),
            #                (0., 1., 0., 0.),
            #                (0., 0., 0., 1.)))

            from scipy.linalg import inv
            aff = inv(aff)

        else:
            raise ValueError("Bad registration procedure type")

        return aff

    def run_pipeline(self):
        #setup
        if self.subjects_dir is None or self.subjects_dir=='':
            self.subjects_dir = os.environ['SUBJECTS_DIR']
        if self.subject is None or self.subject=='':
            self.subject = os.environ['SUBJECT']

        #get rid of any existing grid changes
        self._commit_grid_changes()
        self._grids = {}
        self._grid_types = {}
        
        #manually handle property


        self._electrodes = []
        self._all_electrodes = {}
        self._unsorted_electrodes = {}
        self._sorted_electrodes = {}
        self._interpolated_electrodes = {}

        self._iso_to_surf_map = {}
        self._surf_to_iso_map = {}

        self._snapping_completed = False
        self._noise_hidden = False
        self._visualization_ready = False

        #pipeline
        from . import pipeline as pipe
        
        #if self.use_ct_mask:
        if False:
            ct_mask = pipe.create_brainmask_in_ctspace(self.ct_scan,
                subjects_dir=self.subjects_dir, subject=self.subject,
                overwrite=self.overwrite_xfms)
        else:
            ct_mask = None

        self._electrodes = pipe.remove_large_negative_values_from_ct(
            self.ct_scan, subjects_dir=self.subjects_dir, 
            subject=self.subject)

        self._electrodes = pipe.identify_electrodes_in_ctspace(
            self.ct_scan, mask=ct_mask, threshold=self.ct_threshold,
            use_erosion=(not self.disable_erosion),
            isotropization_type=self.isotropize,
            iso_vector_override=self.isotropization_override) 

        pipe.linearly_transform_electrodes_to_isotropic_coordinate_space(
            self._electrodes, self.ct_scan, 
            isotropization_direction_off = 'copy_to_iso',
            isotropization_direction_on = 'deisotropize',
            isotropization_strategy = self.isotropize,
            iso_vector_override=self.isotropization_override)

        self._electrodes = np.unique(self._electrodes).tolist()

        #I considered allowing the user to manually specify a different
        #registration but we don't currently do this
        aff = self.acquire_affine()

        pipe.create_dural_surface(subjects_dir=self.subjects_dir, 
            subject=self.subject)

        # make the electrodes translated way outside of the brain go away
        # and not be included in the classification at all
        pipe.translate_electrodes_to_surface_space(
            self._electrodes, aff, subjects_dir=self.subjects_dir,
            subject=self.subject)

        if self.use_ct_mask:
            print('eliminating extracranial noise')
            removals=(
                pipe.identify_extracranial_electrodes_in_freesurfer_space(
                self._electrodes, 
                dilation_iterations=self.dilation_iterations,
                subjects_dir=self.subjects_dir, subject=self.subject))

            print('removed %i electrodes' % len(removals))

            for e in removals:
                self._electrodes.remove(e)

        #initial sorting
        #self._grids, self._colors = pipe.classify_electrodes(
        try:
            self._colors, self._grid_geom, self._grids, self._color_scheme = (
                pipe.classify_electrodes(self._electrodes,
                                         self.electrode_geometry,
                                         delta = self.delta,
                                         epsilon = self.epsilon,
                                         rho = self.rho,
                                         rho_strict = self.rho_strict,
                                         rho_loose = self.rho_loose,
                                         crit_pct = self.critical_percentage
                                        ))
        except ValueError as e:
            error_dialog(str(e))
            raise

        # add grid labels to electrodes
        for key in self._grids:
            for elec in self._grids[key]:
                elec.grid_name = key

        # add interpolated points to overall list
        interps = []
        for key in self._grids:
            for elec in self._grids[key]:
                if elec.is_interpolation:
                    interps.append(elec)

        self._electrodes.extend(interps)

        # add isotropization for interpolated points
        pipe.linearly_transform_electrodes_to_isotropic_coordinate_space(
            interps, self.ct_scan,
            isotropization_direction_off = 'copy_to_ct',
            isotropization_direction_on = 'deisotropize',
            isotropization_strategy = self.isotropize,
            iso_vector_override = self.isotropization_override)

        # translate any new electrodes to surface space as well
        pipe.translate_electrodes_to_surface_space(
            self._electrodes, aff, subjects_dir=self.subjects_dir,
            subject=self.subject)

        # make the electrodes translated way outside of the brain go away
        if ct_mask:
            removals=(
                pipe.identify_extracranial_electrodes_in_freesurfer_space(
                    self._electrodes))

            for e in removals:
                self._electrodes.remove(e)
            
        # previously we snapped here

        # Store the sorted/interpolated points in separate maps for access
        for key in self._grids:
            for elec in self._grids[key]:
                if elec.is_interpolation:
                    self._interpolated_electrodes[ intize(elec.asiso()) ]=elec
                else:
                    self._sorted_electrodes[ intize( elec.asiso()) ] = elec

                #save each electrode's grid identity
                self._iso_to_grid_ident_map[ intize( elec.asiso()) ] = key


        #set the grid type to be subdural
        #any depth grids will not have been created by now since they have
        #a geometry
        for key in self._grids:
            self._grid_types[key] = 'subdural'

        # store the unsorted points in a separate map for access
        for elec in self._electrodes:
            is_sorted = False
            for key in self._grids:
                if is_sorted:
                    break
                for elec_other in self._grids[key]:
                    if elec is elec_other:
                        is_sorted = True
                        break
            if not is_sorted:
                self._unsorted_electrodes[ intize(elec.asiso()) ] = elec

                self._iso_to_grid_ident_map[intize(elec.asiso())] = 'unsorted'

        self._all_electrodes.update(self._interpolated_electrodes)
        self._all_electrodes.update(self._unsorted_electrodes)
        self._all_electrodes.update(self._sorted_electrodes)
    
        #save mapping from ct to surface coords
        for elec in list(self._all_electrodes.values()):
            #snapping is never completed by this point anymore
            surf_coord = elec.asras()

            self._iso_to_surf_map[ intize(elec.asiso()) ] = surf_coord
            self._surf_to_iso_map[ intize(surf_coord) ] = elec.asiso()

        #manually trigger a change to grid_named_objects property
        #using an unlikely grid name
        #self._grids['test rice-a-roni'] = []
        #del self._grids['test rice-a-roni']

        self.interactive_mode_displayer.name_holders = (
            self._create_default_name_holders())
        
        self.interactive_mode_displayer.interactive_mode = (
            self.interactive_mode_displayer.name_holders[0])

        self._cursor_tracker = None

        #manually add the new grids to grid_named_objects
        #for key in self._grids:
            #self._grid_named_objects.append( self._new_grid_name_holder(key))
        #    self.interactive_mode_displayer.name_holders.append(
        #        self._new_grid_name_holder(key))

        #fire visualization events
        self._visualization_ready = True
        self._rebuild_vizpanel_event = True
        self._rebuild_guipanel_event = True

    def get_next_color(self):
        color = next(self._color_scheme)
        while color in self._colors:
            color = next(self._color_scheme)
        return color

    def add_grid(self):
        name = 'usergrid%s'%gensym()

        #force self._grids to update (GUI depends on cached property)
#        temp_grids = self._grids
#        self._grids = {}
#        self._grids.update(temp_grids)

        #geometry and color data should be defined first so that when grids
        #grids is updated the GUI does not error out looking for this info
        self._grid_geom[name] = 'user-defined'
        self._colors[name] = self.get_next_color()
        self._grid_types[name] = 'depth'

        #testing GUI update bug
        #temp_grids = self._grids.copy()
        #temp_grids[name] = []
        #self._grids = temp_grids

        self._grids[name] = []
        #self._grid_named_objects = self._get__grid_named_objects()
        #self.interactive_mode = self._grid_named_objects[0]
        #self._grid_named_objects.append( self._new_grid_name_holder( name ))
        self.interactive_mode_displayer.name_holders.append(
            self._new_grid_name_holder(name))
        
        self._update_glyph_lut_event = True

        self._rebuild_interactive_mode_displayer(previous_holder=
            self.interactive_mode_displayer.interactive_mode)

    def add_electrode_to_grid(self, elec, target):
        self._grids[target].append(elec)

        self._iso_to_surf_map[intize(elec.asiso())] = elec.asras()
        self._surf_to_iso_map[intize(elec.asras())] = elec.asiso()

        self._iso_to_grid_ident_map[ intize(elec.asiso()) ] = target
        self._interpolated_electrodes[ intize(elec.asiso()) ] = elec
        self._all_electrodes[ intize(elec.asiso()) ] = elec

        self._rebuild_vizpanel_event = True
        elec.special_name = ''

    def change_single_glyph(self, xyz, elec, target, current_key):
        if elec in self._grids[target]:
            if intize(xyz) in self._points_to_unsorted:
                del self._points_to_unsorted[ intize(xyz) ]
                elec.grid_transition_to = ''
                self._new_glyph_color = list(self._colors.keys()).index(current_key)
            else:
                self._points_to_unsorted[ intize(xyz) ] = elec
                elec.grid_transition_to = 'unsorted'
                self._new_glyph_color = list(self._colors.keys()).index('unsorted')
        else:
            if intize(xyz) in self._points_to_cur_grid:
                del self._points_to_cur_grid[ intize(xyz) ]
                elec.grid_transition_to = ''
                self._new_glyph_color = list(self._colors.keys()).index(current_key)
            else:
                self._points_to_cur_grid[ intize(xyz) ] = elec
                elec.grid_transition_to = target
                self._new_glyph_color = list(self._colors.keys()).index(target)

        self._single_glyph_to_recolor = xyz
        self._update_single_glyph_event = True

    def hide_noise(self):
        self._commit_grid_changes()

        if self._noise_hidden:
            #self._update_glyph_lut_event = True
            self._noise_hidden = False
            self._rebuild_vizpanel_event = True
        else:
            #self._hide_noise_event = True
            self._noise_hidden = True
            self._rebuild_vizpanel_event = True

        self._draw_event = True

#    def open_registration_window(self):
#        cur_grid = self.interactive_mode_displayer.interactive_mode
#        #TODO WHY IS REGISTRATION DONE BY GRIDS? Shouldnt we adjust the
#        #entire image registration?
#        if cur_grid is None:
#            error_dialog('Registration is done gridwise')
#            return
#        if cur_grid.name in ('', 'unsorted'):
#            error_dialog('Regsitration is done gridwise')
#            return
#
#        from utils import RegistrationAdjustmentWindow
#        self.raw = RegistrationAdjustmentWindow(
#            model = self,
#            cur_grid = cur_grid.name)
#        self.raw.edit_traits()
#
#    # this an operation for the manual registration window
#    def reorient_glyph(self, target, matrix):
#        self._commit_grid_changes()
#
#        old_locs = map(lambda x:x.asras(), self._grids[target])
#
#        from geometry import apply_affine
#        new_locs = apply_affine(old_locs, matrix)
#
#        for oloc, nloc in zip(old_locs, new_locs):
#            self._new_ras_positions[tuple(oloc)] = tuple(nloc)
#
#        self._interactive_mode_snapshot = target
#
#        #draw the changes, no underlying state has yet been altered
#        self._reorient_glyph_event = True
#
#        #for unknown reason the operation causes the LUT to be destroyed
#        #we can recreate it easily
#        self._update_glyph_lut_event = True
#
#        #import pdb
#        #pdb.set_trace()
#
#        for oloc,nloc,elec in zip(old_locs, new_locs, self._grids[target]):
#
#            if tuple(oloc) == tuple(nloc):
#                continue
#
#            elec.surf_coords = tuple(nloc)
#            self._iso_to_surf_map[elec.asiso()] = tuple(nloc)
#            del self._surf_to_iso_map[tuple(oloc)]
#            self._surf_to_iso_map[tuple(nloc)] = elec.asiso()
#
#    def fit_changes(self):
#        #maybe this should be a different call which evaluates a single
#        #grid
#
#        #ceurrently we dont use this
#        _, _, self._grids = pipe.classify_electrodes(
#            self._electrodes, self.electrode_geometry,
#            fixed_points=self._grids.values())

    def examine_electrodes(self):
        self._commit_grid_changes()
        cur_grid = self.interactive_mode_displayer.interactive_mode

        if cur_grid is None:
            error_dialog('Select a grid to assign labels')
            return
        if cur_grid.name in ('','unsorted'):
            error_dialog('Select a grid to assign labels')
            return

        if cur_grid.name in self.ews:
            error_dialog('This window is already open')
            return

        #from utils import AutomatedAssignmentWindow
        from .electrode import ElectrodeWindow
        #self.ew = AutomatedAssignmentWindow(
        grid_type = self._grid_types[cur_grid.name]
        ew = ElectrodeWindow(
            model = self,
            cur_grid = cur_grid.name,
            name_stem = cur_grid.name,
            electrodes = self._grids[cur_grid.name],
            grid_type = grid_type,
            naming_convention = ('line' if grid_type == 'depth' 
                else 'grid_serial'),
            )
        self.ews[cur_grid.name] = ew
        ew.edit_traits()

    def open_add_label_window(self):
        if self.alw is None:
            from .utils import AddLabelsWindow
            self.alw = AddLabelsWindow(model=self)
        self.alw.edit_traits()

#    def reconstruct_all_geometry(self):
#        import pipeline as pipe
#
#        for key in self._grids:
#            pipe.classify_single_fixed_grid(key, self._grids, self._grid_geom,
#                self._colors, 
#                delta=self.delta, 
#                epsilon=self.epsilon,
#                rho=self.rho, 
#                rho_loose=self.rho_loose,
#                rho_strict=self.rho_strict)

    def snap_all(self):
        self._commit_grid_changes()

        from . import pipeline as pipe

        snappable_electrodes = []

        for key in list(self._grids.keys()):
            if self._grid_types[key] != 'subdural':
                continue

            snappable_electrodes.extend(self._grids[key])

        if len(snappable_electrodes) == 0:
            error_dialog("Found no subdural electrodes to snap")
            return

        pipe.snap_electrodes_to_surface(
            snappable_electrodes, subjects_dir=self.subjects_dir,
            subject=self.subject, 
            max_steps=self.sa_steps_total,
            giveup_steps=self.sa_steps_break,
            init_temp=self.sa_init_temp,
            temperature_exponent=self.sa_exp,
            )

        self._snapping_completed = True

        #update CT to surf mappings for clickability
        for key in list(self._grids.keys()):
            if self._grid_types[key] == 'subdural':
                for elec in self._grids[key]: 
                    
                    # we snap to the pial surface in terms of returning results
                    # which is pial_coords. but we use the dural surface
                    # coords *specifically for visualization only*.
                    # so the map which is for visualization effects translating
                    # between the two has to use the dural coordinate.
    
                    #snap_coord = elec.astuple()
                    snap_coord = elec.snap_coords
                    surf_coord = elec.asras()
                    iso_coord = elec.asiso()

                    self._iso_to_surf_map[ intize(iso_coord) ] = snap_coord
                    self._surf_to_iso_map[ intize(snap_coord) ] = iso_coord
                    #TODO manage collisions in ct_to_surf mapping
                    
                    # could have been snapped before (in which case we should also get rid of the previous snapped coord in this map)
                    try:
                        del self._surf_to_iso_map[ intize(surf_coord) ]
                    except KeyError:
                        pass

        # we can update the visualization now
        self._rebuild_vizpanel_event = True
    
    def construct_panel2d(self):
        def build_panel():
            from . import panel2d
            self.panel2d = pd = panel2d.TwoDimensionalPanel()
            #pd.on_trait_change( self._create_new_electrode, 
            #    'add_electrode_event')
            pd.load_img(os.path.join(get_subjects_dir(subject=self.subject,
                subjects_dir=self.subjects_dir), 'mri', 'orig.mgz'), 
                image_name='t1')
            pd.load_img(self.ct_scan, image_name='ct')    

        if self.panel2d is None:
            build_panel()
        elif len(self.panel2d.images) < 2:
            expected_images = 0
            if self.subject and self.subjects_dir:
                expected_images += 1
            if self.ct_scan:
                expected_images += 1
            if expected_images == 2:
                build_panel()

        return self.panel2d

    def move_electrode(self, elec, new_coords, in_ras=False,
        as_postprocessing=False):
        '''
        Move the electrode to fix errors, or for postprocessing visualization.
        
        This is done when the user goes into the electrode window, selects an
        electrode, and clicks on the menu item "move this electrode."

        If done as postprocessing, ROIs are not reset and snapped coordinates
        are updated to the new values from the RAS (even if these RAS values 
        were just calculated

        This eliminates the effect of snapping and usually should be done
        before snapping, also resets ROIs.
        '''

        #remove old electrode position from dictionary data structures
        del self._iso_to_surf_map[ intize(elec.asiso()) ]
        del self._surf_to_iso_map[ intize(elec.asras()) ]

        target_grid = self._iso_to_grid_ident_map[ intize(elec.asiso()) ]
        del self._iso_to_grid_ident_map[ intize(elec.asiso()) ]
        del self._all_electrodes[ intize(elec.asiso()) ]

        if not in_ras:
            elec.ct_coords = new_coords
            aff = self.acquire_affine()

            from . import pipeline as pipe
            pipe.translate_electrodes_to_surface_space( [elec], aff,
                subjects_dir=self.subjects_dir, subject=self.subject) 
        else:
            #leave CT alone
            elec.surf_coords = new_coords

        new_ras_coords = elec.asras()

        if as_postprocessing:
            if self._snapping_completed:
                elec.snap_coords = new_ras_coords
                elec.pial_coords = new_ras_coords
        else:
            self._snapping_completed = False
            elec.snap_coords = None
            elec.pial_coords = None
            elec.roi_list = []

        #repopulate dictionaries with new updated electrode
        self._iso_to_surf_map[ intize( elec.asiso()) ] = elec.asras()
        self._surf_to_iso_map[ intize( elec.asras()) ] = elec.asiso()

        self._iso_to_grid_ident_map[ intize( elec.asiso() )] = target_grid
        self._all_electrodes[ intize(elec.asiso()) ] = elec

        self._rebuild_vizpanel_event = True

    @on_trait_change('panel2d:add_electrode_event')
    def _create_new_electrode(self):
        self._commit_grid_changes()

        if len(self._all_electrodes) == 0:
            error_dialog('No electrodes loaded')
            return

        pd = self.panel2d
        image_name = pd.currently_showing.name
        px, py, pz = pd.cursor

        #px,py,pz,_ = pd.pins[image_name][pd.current_pin]

        from .electrode import Electrode
        elec = Electrode(ct_coords = (px,py,pz))

        if image_name == 't1':
            #rx,ry,rz = pd.map_cursor((px,py,pz), pd.images['t1'][2])
            error_dialog('Adding electrodes only allowed from CT reference')
            return
    
        elif image_name == 'ct':
            aff = self.acquire_affine()
            from . import pipeline as pipe

            pipe.translate_electrodes_to_surface_space( [elec], aff,
                subjects_dir=self.subjects_dir, subject=self.subject)

            pipe.linearly_transform_electrodes_to_isotropic_coordinate_space(
                [elec], self.ct_scan,
                isotropization_direction_off = 'copy_to_iso',
                isotropization_direction_on = 'isotropize',
                isotropization_strategy = self.isotropize,
                iso_vector_override = self.isotropization_override)
        else:
            raise ValueError("Internal error: bad image type")

        #self._grids['unsorted'] = elec
        #self._grids['unsorted'].append(elec)

        self._iso_to_surf_map[ intize(elec.asiso()) ] = elec.asras()
        self._surf_to_iso_map[ intize(elec.asras()) ] = elec.asiso()

        self._iso_to_grid_ident_map[ intize(elec.asiso()) ] = 'unsorted'
        self._all_electrodes[ intize(elec.asiso()) ] = elec
        self._unsorted_electrodes[ intize( elec.asiso()) ] = elec

        self._rebuild_vizpanel_event = True

    @on_trait_change('panel2d:untrack_cursor_event')
    def _remove_tracked_cursor(self):
        self._cursor_tracker = None

    @on_trait_change('panel2d:track_cursor_event')
    def _add_tracked_cursor(self):
        self._commit_grid_changes()

        #if there is already a tracker and the user hits this button, it
        #becomes an undo function. remove the existing tracker and exit
        if self._cursor_tracker is not None:
            self._cursor_tracker = None
            self._rebuild_vizpanel_event = True
            return

        pd = self.panel2d
        image_name = pd.currently_showing.name
        px, py, pz = pd.cursor

        if image_name == 't1':
            error_dialog('Cursor can only be tracked from CT reference')
            return

        elif image_name == 'ct':
            elec = Electrode(ct_coords = (px,py,pz))
            self._cursor_tracker = elec
            
            aff = self.acquire_affine()
            from . import pipeline as pipe
            pipe.linearly_transform_electrodes_to_isotropic_coordinate_space(
                [elec], self.ct_scan,
                isotropization_direction_off = 'copy_to_iso',
                isotropization_direction_on = 'isotropize',
                isotropization_strategy = self.isotropize,
                iso_vector_override = self.isotropization_override)

            pipe.translate_electrodes_to_surface_space( [elec], aff,
                subjects_dir=self.subjects_dir, subject=self.subject)

        else:
            raise ValueError("Internal error: bad image type")

        self._rebuild_vizpanel_event = True

    @on_trait_change('panel2d:panel2d_closed_event')
    def _removed_tracked_cursor(self):
        self._commit_grid_changes()

        if self._cursor_tracker is not None:
            self._rebuild_vizpanel_event = True

    def _ask_user_for_savefile(self, title=None):
        from .utils import ask_user_for_savefile
        return ask_user_for_savefile(title=title)

    def _ask_user_for_loadfile(self, title=None):
        from .utils import ask_user_for_loadfile
        return ask_user_for_loadfile(title=title)

    def _save_to_pickle(self, savefile):
        self._commit_grid_changes()

        from pickle import dump
        try:
            with open(savefile, 'wb') as fd:
                dump(self, fd)
                print('Saved pickle file {0}'.format(savefile))
        except (KeyError, AttributeError, TypeError) as e:
            error_dialog('Failed to save pickle file {0}'.format(savefile))
            print(e)

    def _load_from_pickle(self, loadfile):
        from pickle import load
        #from restricted_pickler import restricted_load
        try:
            with open(loadfile, "rb") as fd:
                pickle_model = load(fd)
            #pickle_model = restricted_load(loadfile)
        except (KeyError, AttributeError, TypeError) as e:
            error_dialog('Failed to load ElectrodePositionsModel object\n'
                'from provided pickle file {0}.\n\n'
                'Are you sure that was a correct pickle file?'
                .format(loadfile))
            print(e)
            return

        #automatically assign pickle to model (old)
        #self.model = pickle_model

        self.__dict__.update(pickle_model.__dict__)

            #manually assign every attribute
#            for field in pickle_model.__dict__:
#                try:
#                    setattr(self.model, field, getattr(pickle_model, field))
#                    #self.model.field = pickle_model.field
#                except Exception as e:

        #color scheme is a generator and therefore can't be imported properly
        #but we can recreate the correct generator state easily enough
        from .utils import get_default_color_scheme
        colors = get_default_color_scheme()
        #get rid of the right number of colors
        for i in range(len(self._grids)):
            next(colors)

        self._color_scheme = colors

        #the trait notifiers set up on the displayer are not reproduced
        #after pickling since those are not defined in the NameHolder
        #initializer but by private method _new_grid_name_holder(grid_name)
        #maybe they could be defined in the initializer instead but whatever
        #here we just manually call our tools to rebuild the displayer safely


        self._rebuild_interactive_mode_displayer()

        self._rebuild_guipanel_event = True
        self._rebuild_vizpanel_event = True

    def add_annotation(self, annot_name, hemi='both', border=True, opacity=1.):
        self._label_file = annot_name
        self._label_borders = border
        self._label_opacity = opacity
        self._label_hemi = hemi
        self._add_annotation_event = True

    def add_label(self, label_file, border=True, opacity=1., color='blue'):
        self._label_file = label_file
        self._label_borders = border
        self._label_opacity = opacity
        self._label_color = color
        self._add_label_event = True

    def remove_labels(self):
        self._remove_labels_event = True

    def get_electrodes_from_grid(self, target=None, electrodes=None):
        if target is None:
            self._commit_grid_changes()
            if self.interactive_mode_displayer.interactive_mode is None:
                print("select a grid to save labels from")
                return
            target = self.interactive_mode_displayer.interactive_mode.name
            if target in ('unsorted',):
                print("select a grid to save labels from")
                return

        #for now only save label files for the current grid, may change
        #in principle this is not what we want

        # get electrodes from current state if not passed as argument
        key = target
        if electrodes is None:
            key = self.interactive_mode_displayer.interactive_mode.name
            electrodes = self._grids[key]

        return sorted(electrodes)

    def get_electrodes_all(self):
        return sorted([e for e in list(self._all_electrodes.values()) if e.grid_name != 'unsorted'])

class ExtractionRegistrationSortingPanel(HasTraits):
    model = Instance(ElectrodePositionsModel)

    ct_threshold = DelegatesTo('model')
    critical_percentage = DelegatesTo('model')
    delta = DelegatesTo('model')
    epsilon = DelegatesTo('model')
    rho = DelegatesTo('model')
    rho_loose = DelegatesTo('model')
    rho_strict = DelegatesTo('model')
    delta_recon = DelegatesTo('model')
    epsilon_recon = DelegatesTo('model')
    rho_recon = DelegatesTo('model')
    rho_loose_recon = DelegatesTo('model')
    rho_strict_recon = DelegatesTo('model')
    #visualize_in_ctspace = DelegatesTo('model')
    sa_steps_break = DelegatesTo('model')
    sa_steps_total = DelegatesTo('model')
    sa_init_temp = DelegatesTo('model')
    sa_exp = DelegatesTo('model')
    deformation_constant = DelegatesTo('model')
    use_ct_mask = DelegatesTo('model')
    disable_erosion = DelegatesTo('model')
    overwrite_xfms = DelegatesTo('model')
    registration_procedure = DelegatesTo('model')
    registration_algorithm = DelegatesTo('model')
    shapereg_slice_diff = DelegatesTo('model')
    zoom_factor_override = DelegatesTo('model')
    dilation_iterations = DelegatesTo('model')
    isotropize = DelegatesTo('model')
    isotropization_override = DelegatesTo('model')

    traits_view = View(
        Group(
        HGroup(
        VGroup(
            Label('The threshold above which electrode clusters will be\n'
                'extracted from the CT image'),
            Item('ct_threshold'),
            Label('Weight given to the deformation term in the snapping\n'
                'algorithm, reduce if snapping error is very high.'),
            Item('deformation_constant'),
            Label('Mask extracranial noise'),
            HGroup(
                Item('use_ct_mask', show_label=False),
                Item('dilation_iterations', show_label=True, 
                    label='iterations', enabled_when='use_ct_mask'),
            ),
            Label('Overwrite existing transformations'),
            Item('overwrite_xfms'),
            Label('Disable binary erosion procedure to reduce CT noise'),
            Item('disable_erosion'),
            HGroup(
                VGroup(
                Label('Type of registration'),
                Item('registration_procedure', show_label=False),
                ),
                VGroup(
                Label('Override zoom factor'),
                Item('zoom_factor_override', editor =CSVListEditor(),
                enabled_when='registration_procedure==\'experimental shape '
                'correction\'', show_label=False),
                ),
            ),
            Label('Registration cost function'),
            Item('registration_algorithm', show_label=False),
            Label('Convert electrode locations to isotropic coordinate '
                'space\nbefore sorting'),
            HGroup(
                Item('isotropize', show_label=False),
                Item('isotropization_override', editor=CSVListEditor(),
                    enabled_when='isotropize==\'Manual override\'',
                    show_label=False),
            ),
        show_labels=False),
        VGroup(
            Label('The percentage of electrodes to find in sorting'),
            Item('critical_percentage'),
            Label('Delta controls the distance between electrodes. That is,\n'
                'electrode distances must be between c*(1-d) and c*(1+d),\n'
                'where c is an estimate of the correct distance.'),
            Item('delta'),
            Label('Epsilon controls the tolerance of the initial angle\n'
                'difference from 90 degrees (or 180 degrees in rare cases).'),
            Item('epsilon'),
            Label('Rho controls the maximum allowable discrepancy between\n'
                'angles relative to their position to already fitted\n'
                'electrodes mostly as the difference |rho-90|.\n'
                'Rho_strict and Rho_loose are used in similar cases,\n' 
                'which demand slightly greater or smaller constraints.'),
            Item('rho'),
            Item('rho_strict'),
            Item('rho_loose'),
            Label('Simulated annealing parameters'),
            HGroup(
                Item('sa_steps_break', label='steps before convergence'),
                Item('sa_steps_total', label='steps total'),
            ),
            HGroup(
                Item('sa_init_temp', label='initial temperature'),
                Item('sa_exp', label='exponential term'),
            ),
        ),
        ),
        ),
        
    title='Adjust extraction registration and sorting parameters',
    buttons=OKCancelButtons,
    )

class VisualizationsOutputsPanel(HasTraits):
    model = Instance(ElectrodePositionsModel)

    roi_parcellation = DelegatesTo('model')
    roi_error_radius = DelegatesTo('model')

    coronal_dpi = DelegatesTo('model')
    coronal_size = DelegatesTo('model')

    traits_view = View(
        HGroup(
        VGroup(
            Label('Identification of ROIs near electrodes'),
            Item('roi_parcellation'),
            Item('roi_error_radius'),
        ),
        VGroup(
            Label('Coronal slices'),
            Item('coronal_dpi', label='dpi'),
            Item('coronal_size', label='size', editor=CSVListEditor()),
        ),
        ),
        
    title='Adjust clinical visualization and output parameters',
    buttons=OKCancelButtons,
    )

class SurfaceVisualizerPanel(HasTraits):
    scene = Instance(MlabSceneModel,())
    model = Instance(ElectrodePositionsModel)

    subject = DelegatesTo('model')
    subjects_dir = DelegatesTo('model')
    _colors = DelegatesTo('model')

    _grids = DelegatesTo('model')
    #interactive_mode = DelegatesTo('model')
    interactive_mode_displayer = DelegatesTo('model')

    _points_to_unsorted = DelegatesTo('model')
    _points_to_cur_grid = DelegatesTo('model')

    _all_electrodes = DelegatesTo('model')
    _unsorted_electrodes = DelegatesTo('model')
    _iso_to_surf_map = DelegatesTo('model')
    _surf_to_iso_map = DelegatesTo('model')
    _grid_types = DelegatesTo('model')

    _cursor_tracker = DelegatesTo('model')

    visualize_in_ctspace = Bool(False)
    _viz_coordtype = Property#(depends_on='visualize_in_ctspace')
    def _get__viz_coordtype(self):
        if self.visualize_in_ctspace:
            #return 'ct_coords'
            return 'iso_coords'
        elif self.model._snapping_completed:
            #plot points on dural surface better for visualization than pial
            return 'snap_coords'
            #return 'pial_coords'
        else:
            return 'surf_coords'

    brain = Any
    gs_glyphs = Dict
    tracking_glyph = Any

    _lh_pysurfer_offset = Float
    _rh_pysurfer_offset = Float
    def __lh_pysurfer_offset_default(self):
        if self.visualize_in_ctspace:
            return 0
        lh_pia, _ = nib.freesurfer.read_geometry(
            os.path.join(self.subjects_dir, self.subject, 'surf', 'lh.pial'))
        return np.max(lh_pia[:, 0])
    def __rh_pysurfer_offset_default(self):
        if self.visualize_in_ctspace:
            return 0
        rh_pia, _ = nib.freesurfer.read_geometry(
            os.path.join(self.subjects_dir, self.subject, 'surf', 'rh.pial'))
        return np.min(rh_pia[:, 0])

    traits_view = View(
        Item('scene', editor=SceneEditor(scene_class=MayaviScene),
            show_label=False, resizable=True),
        height=500, width=500, resizable=True)

    def __init__(self, model, **kwargs):
        super(SurfaceVisualizerPanel, self).__init__(**kwargs)
        self.model = model

    @on_trait_change('scene:activated')
    def setup(self):
        #import pdb
        #pdb.set_trace()
        if self.model._visualization_ready:
            #print self.model._cursor_tracker
            #print self._cursor_tracker
            self.show_grids_on_surface()

    def show_grids_on_surface(self):

        from mayavi import mlab
        #mlab.clf(figure = self.scene.mayavi_scene)

        #there is a bug in mlab.clf which causes the callbacks to become
        #disconnected in such a way that they cannot be reattached to the
        #scene. I tracked this bug to the VTK picker before giving up.

        #To avoid this, we use a workaround -- discard the scene every
        #single time we want to use mlab.clf and replace it with an
        #entirely new MlabSceneModel instance. This has the added benefit
        #of (according to my tests) avoiding memory leaks.

        #from utils import clear_scene
        #clear_scene(self.scene.mayavi_scene)

        from .color_utils import set_discrete_lut

        import surfer
        if not self.visualize_in_ctspace:
            brain = self.brain = surfer.Brain( 
                self.subject, subjects_dir=self.subjects_dir,
                surf='pial', hemi='both',
                figure=self.scene.mayavi_scene)

            brain.toggle_toolbars(True)

            #set the surface unpickable
            for srf in brain.brains:
                srf._geo_surf.actor.actor.pickable=False
                srf._geo_surf.actor.property.opacity = (
                    self.model.surface_opacity)

            scale_factor = 3.
        else:
            scale_factor = 5.

        def fix_elec( elec, coordtype=None ):
            init_coord = np.array(getattr(elec, coordtype))

            if init_coord[0] < 0: 
                init_coord[0] -= self._lh_pysurfer_offset
            else:
                init_coord[0] -= self._rh_pysurfer_offset
            return init_coord

        unsorted_coordtype = (self._viz_coordtype if 
            self._viz_coordtype not in ('pial_coords', 'snap_coords')
            else 'surf_coords')
        
        #from PyQt5.QtCore import pyqtRemoveInputHook
        #import pdb
        #pyqtRemoveInputHook()
        #pdb.set_trace()

        #unsorted
        if not self.model._noise_hidden:
        
            unsorted_elecs = list(map(
                partial(fix_elec, coordtype=unsorted_coordtype),
                 list(self._unsorted_electrodes.values()) ))
            self.gs_glyphs['unsorted'] = glyph = virtual_points3d( 
                unsorted_elecs, scale_factor=scale_factor, name='unsorted',
                figure=self.scene.mayavi_scene, 
                color=self._colors['unsorted'])  

            set_discrete_lut(glyph, list(self._colors.values()))
            glyph.mlab_source.dataset.point_data.scalars=(
                np.zeros(len(unsorted_elecs)))

        #grids
        for i,key in enumerate(self._grids):

            grid_coordtype = (self._viz_coordtype if
                (self._viz_coordtype not in ('snap_coords', 'pial_coords') or
                 self._grid_types[key]=='subdural') else 
                'surf_coords')

            grid_elecs = list(map(
                partial(fix_elec, coordtype=grid_coordtype),
                self._grids[key]))
                        
            if len(grid_elecs)==0:
                continue

            self.gs_glyphs[key] = glyph = virtual_points3d(grid_elecs,
                scale_factor=scale_factor, color=self._colors[key], 
                name=key, figure=self.scene.mayavi_scene)

            set_discrete_lut(glyph, list(self._colors.values()))
            scalar_color = list(self._colors.keys()).index(key)

            glyph.mlab_source.dataset.point_data.scalars=(
                np.ones(len(self._grids[key])) * scalar_color)

        #panel tracking
        if self._cursor_tracker is not None:
            tracker_elec = fix_elec(self._cursor_tracker,
                coordtype=unsorted_coordtype)

            tracking_scale_factor = 6. if self.visualize_in_ctspace else 4.

            self.tracking_glyph = virtual_points3d( [tracker_elec], 
                scale_factor=tracking_scale_factor, name='tracking_glyph',
                figure=self.scene.mayavi_scene, 
                color=self._colors['selection'])

        #setup the node selection callback
        picker = self.scene.mayavi_scene.on_mouse_pick( self.selectnode_cb )
        picker.tolerance = .02

    def redraw_single_grid(self, key):
        #this function is never called
        #it is always easier to redraw everything, and avoids memory leaks

        from mayavi import mlab
        from .color_utils import set_discrete_lut

        #removing glyphs may cause memory leaks
        self.gs_glyphs[key].remove()

        grid_elecs = list(map((lambda x:getattr(x, self._viz_coordtype)),
            self._grids[key]))

        if len(grid_elecs)==0:
            return

        self.gs_glyphs[key] = glyph = virtual_points3d(grid_elecs,
            scale_factor=scale_factor, color=self._colors[key],
            name=key, figure=self.scene.mayavi_scene)
      
        set_discrete_lut(glyph, list(self._colors.values()))
        scalar_color = list(self._colors.keys()).index(key)

        glyph.mlab_source.dataset.point_data.scalars=(
            np.ones(len(self._grids[key])) * scalar_color)

    def selectnode_cb(self, picker):
        '''
        Callback to move an node into the selected state
        '''
        #import pdb
        #pdb.set_trace()

        if self.interactive_mode_displayer.interactive_mode is None:
            return
        target = self.interactive_mode_displayer.interactive_mode.name
        if target in ('', 'unsorted', 'selection'):
            return

        current_key = None

        for key,nodes in zip(list(self.gs_glyphs.keys()), list(self.gs_glyphs.values())):
            if picker.actor in nodes.actor.actors:
                pt = int(picker.point_id/nodes.glyph.glyph_source.
                    glyph_source.output.points.to_array().shape[0])
                #x,y,z = np.around(nodes.mlab_source.points[pt], 4)
                x,y,z = nodes.mlab_source.points[pt]

                #translate from CT to surf coords if necessary
                if not self.visualize_in_ctspace:
                    if x < 0:
                        x += self._lh_pysurfer_offset
                    else:
                        x += self._rh_pysurfer_offset
                    x,y,z = self._surf_to_iso_map[intize((x,y,z)) ]
                    #x,y,z = self._iso_to_surf_map[intize((x,y,z)) ]

                elec = self._all_electrodes[ intize((x,y,z))]

                current_key = elec.grid_name
                break

        #if the user did not click on anything interesting, do nothing
        if current_key is None:
            return

        self.model.change_single_glyph((x,y,z), elec, target, current_key)
    
    @on_trait_change('model:_update_single_glyph_event')
    def update_single_glyph(self):

        #to be safe
        #self.update_glyph_lut()

        if len(self.gs_glyphs)==0:
            #this callback is still hooked to the SurfaceVisualizerPanel 
            #instance which has been disconnected from the scene
            return

        from .color_utils import change_single_glyph_color
        from mayavi import mlab
        #import pdb
        #pdb.set_trace()
        xyz = self.model._single_glyph_to_recolor
        if not self.visualize_in_ctspace:
            xyz = np.array( self._iso_to_surf_map[ intize(xyz) ] )
            if xyz[0] < 0:
                xyz[0] -= self._lh_pysurfer_offset
            else:
                xyz[0] -= self._rh_pysurfer_offset

        for nodes in list(self.gs_glyphs.values()):
            pt, = np.where( np.all(nodes.mlab_source.points == xyz, axis=1 ))
            if len(pt) > 0:
                break

        if len(pt)==0:
            #why does this error sometimes get thrown when a correct point
            #was clicked and the visualization changes
            raise ValueError('Error in figuring out what point was clicked')

        change_single_glyph_color(nodes, int(pt), self.model._new_glyph_color)

        mlab.draw(figure=self.scene.mayavi_scene)

    @on_trait_change('model:_draw_event')
    def force_render(self):
        from .plotting_utils import force_render as fr
        fr( figure=self.scene.mayavi_scene )

    @on_trait_change('model:_update_glyph_lut_event')
    def update_glyph_lut(self):
        from .color_utils import set_discrete_lut
        for glyph in list(self.gs_glyphs.values()):
            set_discrete_lut(glyph, list(self._colors.values()))

    #not used. has the problem of crescent moon shaped electrodes,
    #transparent electrodes are still occlusive despite being transparent
    @on_trait_change('model:_hide_noise_event')
    def hide_unsorted_electrodes(self):
        from .color_utils import make_transparent
        for glyph in list(self.gs_glyphs.values()):
            make_transparent(glyph, 0) 

    @on_trait_change('model:_add_annotation_event')
    def add_annotation(self):
        if self.visualize_in_ctspace:
            return
        
        if self.brain is None:
            error_dialog("Run pipeline first")
            return

        #for hemi in ('lh', 'rh'):
        if self.model._label_hemi in ('lh','rh'):
            self.brain.add_annotation(self.model._label_file,
                borders=self.model._label_borders,
                alpha=self.model._label_opacity,
                hemi=self.model._label_hemi,
                remove_existing=True)

        else:
            for hemi in ('lh','rh'):
                self.brain.add_annotation(self.model._label_file,
                    borders=self.model._label_borders,
                    alpha=self.model._label_opacity,
                    hemi=hemi,
                    remove_existing=(hemi=='lh'))
        
    @on_trait_change('model:_add_label_event')
    def add_label(self):
        if self.visualize_in_ctspace:
            return

        if self.brain is None:
            error_dialog('Run pipeline first')
            return

        import mne
        from .color_utils import traits2mayavi_color
        self.brain.add_label(self.model._label_file,
            borders=self.model._label_borders,
            alpha=self.model._label_opacity,
            color=traits2mayavi_color(self.model._label_color),
            hemi=mne.read_label(self.model._label_file).hemi )

    @on_trait_change('model:_remove_labels_event')
    def remove_labels(self):
        if self.visualize_in_ctspace or self.brain is None:
            return

        self.brain.remove_labels(hemi='lh')
        self.brain.remove_labels(hemi='rh')
        for annot in self.brain.annot_list:
            annot['surface'].remove()
        self.brain.annot_list=[]
    
    #@on_trait_change('model:_reorient_glyph_event')
    #def update_orientation(self):
        #only do this for surface viz, do not adjust CT electrode positions
    #    if self.visualize_in_ctspace:
    #        return
    #
    #    target = self.model._interactive_mode_snapshot
    #
    #    for glyph in self.gs_glyphs.values():
    #        points = np.array(glyph.mlab_source.dataset.points)
    #
    #        for i,point in enumerate(points):
    #            surf_point = tuple(self.model._surf_to_iso_map[tuple(point)])
    #            if self.model._iso_to_grid_ident_map[surf_point] == target:
    #                points[i] = self.model._new_ras_positions[tuple(point)] 
    #
    #        glyph.mlab_source.dataset.points = points

        #it is necessary to either
            #1. redraw the scene (we are not doing this)
            #2. determine the electrode identities, by iterating through
            #   each glyphs and somehow figuring out the electrode
            #   identity, perhaps by maintaining a table in the model

            #   and then storing and updating the locations of
            #   *all* glyphs

            #   note, all the electrodes we want are in one Grid
            #   but not necessarily one glyph.
        
            #   so it does not make sense to calculate all the electrode
            #   locations and leave them unordered, because we will need to 
            #   replace the positions array for each glyph individually
            #   after we figure out the membership of each electrode

            #   so it is probably best to just to access each electrode
            #   directly from the model to get its updated coordinates

    @on_trait_change('model:surface_opacity')
    def update_surface_opacity(self):
        if not self.visualize_in_ctspace:
            opc = self.model.surface_opacity
            for srf in self.brain.brains:
                srf._geo_surf.actor.property.opacity = opc
            

class InteractivePanel(HasPrivateTraits):
    model = Instance(ElectrodePositionsModel)

    ct_scan = DelegatesTo('model')
    t1_scan = DelegatesTo('model')
    run_pipeline_button = Button('Run pipeline')

    subjects_dir = DelegatesTo('model')
    subject = DelegatesTo('model')
    fsdir_writable = DelegatesTo('model')

    ct_registration = DelegatesTo('model')

    electrode_geometry = DelegatesTo('model')

    _grid_named_objects = DelegatesTo('model')

    #interactive_mode = Instance(NameHolder)
    #interactive_mode = DelegatesTo('model')
    interactive_mode_displayer = DelegatesTo('model')

    add_grid_button = Button('Add new grid')
    add_label_button = Button('Add labels')
    shell = Dict

    save_montage_button = Button('Save montage')
    save_csv_button = Button('Save csv')

    edit_parameters_button = Button('Edit Fitting Parameters')
    hide_noise_button = Button('Hide noise')
    
    reconstruct_vizpanel_button = Button('Rebuild viz')
    examine_electrodes_button = Button('Examine electrodes')
    snap_electrodes_button = Button('Snap electrodes')
    #adjust_registration_button = Button('Adjust registration')
    visualize_ct_button = Button('Examine CT')

    #we retain a reference to easily reference the visualization in the shell
    viz = Instance(SurfaceVisualizerPanel)
    ctviz = Instance(SurfaceVisualizerPanel)

    traits_view = View(
        HGroup(
            VGroup(
                Item('ct_scan'),
                #Item('ct_registration', label='reg matrix\n(optional)')
                #Item('adjust_registration_button', show_label=False),
                #HGroup(
                #    Item('visualize_ct_button', show_label=False),
                #    Item('hide_noise_button', show_label=False),
                #),
            ),
            VGroup(
                Item('electrode_geometry', editor=CustomListEditor(
                    editor=CSVListEditor(), rows=2), ), 
            ), 
            VGroup(
                Item('run_pipeline_button', show_label=False),
                #Item('edit_parameters_button', show_label=False),
                #HGroup(
                #    Item('save_montage_button', show_label=False),
                #    Item('save_csv_button', show_label=False),
                #),
            ),
        ),
        HGroup(
                Item('subjects_dir'),
                Item('subject'),
        ),
        HGroup(
            VGroup(
                #Item('interactive_mode', 
                #    editor=InstanceEditor(name='_grid_named_objects'),
                #    style='custom', label='Edit electrodes\nfrom grid'),
                Item('interactive_mode_displayer',
                    editor=InstanceEditor(), style='custom',
                    label='Edit electrodes\nfrom grid'),
            ),
            VGroup(
                Item('add_grid_button', show_label=False),
                #Item('reconstruct_vizpanel_button', show_label=False),
                #Item('add_label_button', show_label=False)
            #),
            #VGroup(
                Item('examine_electrodes_button', show_label=False),
            #    Item('snap_electrodes_button', show_label=False),
            ),
        ),

                Item('shell', show_label=False, editor=ShellEditor()),

        height=300, width=500, resizable=True
    )

    def __init__(self, model, viz=None, ctviz=None, **kwargs):
        super(InteractivePanel, self).__init__(**kwargs)
        self.model = model
        self.viz = viz
        self.ctviz = ctviz

    def _run_pipeline_button_fired(self):
        self.model.run_pipeline()

    def _add_grid_button_fired(self):
        self.model.add_grid()

    def _find_best_fit_button_fired(self):
        self.model.fit_changes()
#
    def _edit_parameters_button_fired(self):
        ExtractionRegistrationSortingPanel(model=self.model).edit_traits()

    def _hide_noise_button_fired(self):
        self.model.hide_noise() 

    def _add_label_button_fired(self):
        self.model.open_add_label_window()

    def _reconstruct_vizpanel_button_fired(self):
        self.model._reconstruct_vizpanel_event = True

    def _snap_electrodes_button_fired(self):
        self.model.snap_all()

    def _examine_electrodes_button_fired(self):
        self.model.examine_electrodes()

    def _adjust_registration_button_fired(self):
        self.model.open_registration_window()

#    def _visualize_ct_button_fired(self):
#        import panel2d
#        self.model.panel2d = pd = panel2d.TwoDimensionalPanel()
#        #pd.on_trait_change(self.model._create_new_electrode, 
#        #    'add_electrode_event')
#        pd.load_img(self.ct_scan, image_name='ct')
#        pd.edit_traits()

class iEEGCoregistrationFrame(HasTraits):
    model = Instance(ElectrodePositionsModel)
    interactive_panel = Instance(InteractivePanel)
    surface_visualizer_panel = Instance(SurfaceVisualizerPanel)
    ct_visualizer_panel = Instance(SurfaceVisualizerPanel)

    load_pickle_action = Action(name='Load pickle', action='do_load_pickle' )
    save_pickle_action = Action(name='Save pickle', action='do_save_pickle' )
    quit_action = Action(name='Quit', action='do_quit')

    run_pipeline_action = Action( name='Run pipeline', 
        action='do_run_pipeline')
    snap_action = Action( name='Snap electrodes to surface', action='do_snap')
    hide_noise_action = Action( name='Hide noise', action='do_hide_noise')
    examine_ct_action = Action( name='Examine CT', action='do_examine_ct')
    display_pysurfer_labels_action = Action(name='Show labels or annotation',
        action='do_display_pysurfer_labels')

    find_rois_action = Action(name='Find neighboring ROIs', 
        action='do_find_rois')
    coronal_slices_action = Action(name='Save coronal slices (depths)',
        action='do_coronal_slices')
    subdural_opaque_images_action = Action(
        name='Save snapshots (grids)',
        action='do_save_opaque')
    save_montage_action = Action(name='Save montage', 
        action='do_save_montage')
    save_csv_action = Action(name='Save CSV', action='do_save_csv')

    edit_extraction_sorting_parameters_action = Action(
        name = 'Extraction/Registration/Sorting',
        action = 'do_extraction_params')
    edit_visualization_output_parameters_action = Action(
        name = 'Visualization/Output',
        action = 'do_visualization_params')


    traits_view = View(
        Group(
            HGroup(
                Item('surface_visualizer_panel', editor=InstanceEditor(), 
                    style='custom', resizable=True ),
                Item('ct_visualizer_panel', editor=InstanceEditor(),
                    style='custom', resizable=True ),
            show_labels=False, layout='split'),
            Item('interactive_panel', editor=InstanceEditor(), style='custom',
                resizable=True),
        show_labels=False, layout='split'),
        title=('llanfairpwllgwyngyllgogerychwyrndrobwllllantysiliogogogoch is'
            ' nice this time of year'),

        menubar = MenuBar(
            Menu( load_pickle_action,
                  save_pickle_action,
                  #quit_action,
                name = 'File'),
            Menu ( run_pipeline_action,
                   examine_ct_action,
                   hide_noise_action,
                   display_pysurfer_labels_action,
                name = 'Tools'),
            Menu( find_rois_action,
                  snap_action, 
                  coronal_slices_action,
                  subdural_opaque_images_action,
                  save_montage_action,
                  save_csv_action,
                name = 'Electrode'),
            Menu ( edit_extraction_sorting_parameters_action,
                   edit_visualization_output_parameters_action,
                name = 'Settings'),
        ),

        height=800, width=800, resizable=True
    )

    def __init__(self, **kwargs):
        super(iEEGCoregistrationFrame, self).__init__(**kwargs)
        model = self.model = ElectrodePositionsModel()
        self.surface_visualizer_panel = SurfaceVisualizerPanel(model)
        self.ct_visualizer_panel = SurfaceVisualizerPanel(model,
            visualize_in_ctspace=True)
        self.interactive_panel = InteractivePanel(model,
            viz=self.surface_visualizer_panel,
            ctviz=self.ct_visualizer_panel)

    def do_save_pickle(self):
        savefile = self.model._ask_user_for_savefile(title='save pkl file')
        self.model._save_to_pickle(savefile)

    def do_load_pickle(self):
        loadfile = self.model._ask_user_for_loadfile(title='load pkl file')
        self.model._load_from_pickle(loadfile)
    
    #def do_quit(self):
    #    sys.exit(0)
    # user can quit with window manager. makes less likely to misclick and
    # exit accidentally

    def do_run_pipeline(self):
        self.model.run_pipeline()

    def do_snap(self):
        self.model.snap_all()

    def do_examine_ct(self):
        from . import panel2d

        pd = self.model.construct_panel2d()

        #self.model.panel2d = pd = panel2d.TwoDimensionalPanel()
        #pd.load_img(self.model.ct_scan, image_name='ct')
        pd.edit_traits()

    def do_hide_noise(self):
        self.model.hide_noise() 

        if self.model._noise_hidden:
            self.hide_noise_action.name = 'Unhide noise'
        else:
            self.hide_noise_action.name = 'Hide noise'

    def do_display_pysurfer_labels(self):
        self.model.open_add_label_window()

    def do_find_rois(self):
        self.model._commit_grid_changes()

        from .electrode_group import get_nearby_rois_all
        get_nearby_rois_all( self.model._grids,
                             subjects_dir=self.model.subjects_dir,
                             subject=self.model.subject,
                             parcellation=self.model.roi_parcellation,
                             error_radius=self.model.roi_error_radius )

    def do_coronal_slices(self):
        self.model._commit_grid_changes()

        from .electrode_group import coronal_slice_all
        coronal_slice_all( self.model._grids,
                           self.model._grid_types,
                           subjects_dir=self.model.subjects_dir,
                           subject=self.model.subject,
                           dpi=self.model.coronal_dpi )

    def do_save_opaque(self):
        self.model._commit_grid_changes()

        savefile = self.model._ask_user_for_savefile()

        if savefile is None:
            return

        self.model.surface_opacity = 1.0
        if not self.model._noise_hidden:
            self.model.hide_noise()

        from .plotting_utils import save_opaque_clinical_sequence
        save_opaque_clinical_sequence( savefile, 
            self.surface_visualizer_panel.scene.mayavi_scene )

        self.model.surface_opacity = 0.4
        if self.model._noise_hidden:
            self.model.hide_noise()

    def do_save_montage(self):
        self.model._commit_grid_changes()

        electrodes = self.model.get_electrodes_all()

        from .electrode_group import save_coordinates
        save_coordinates( electrodes, self.model._grid_types,
            snapping_completed=self.model._snapping_completed,
            file_type='montage')
    
    def do_save_csv(self):
        self.model._commit_grid_changes()

        electrodes = self.model.get_electrodes_all()

        from .electrode_group import save_coordinates
        save_coordinates( electrodes, self.model._grid_types,
            snapping_completed=self.model._snapping_completed,
            file_type='csv')

    def do_extraction_params(self):
        ExtractionRegistrationSortingPanel(model=self.model).edit_traits()

    def do_visualization_params(self):
        VisualizationsOutputsPanel(model=self.model).edit_traits()

    @on_trait_change('model:_rebuild_vizpanel_event')
    def _rebuild_vizpanel(self):

        #throw away old vizpanel listeners to remove annoying error messages
        self.surface_visualizer_panel.model = None
        self.ct_visualizer_panel.model = None

        #create new panels
        self.surface_visualizer_panel = SurfaceVisualizerPanel(self.model)
        self.interactive_panel.viz = self.surface_visualizer_panel

        self.ct_visualizer_panel = SurfaceVisualizerPanel(self.model,
            visualize_in_ctspace=True)
        self.interactive_panel.ctviz = self.ct_visualizer_panel

    @on_trait_change('model:_rebuild_guipanel_event')
    def _rebuild_guipanel(self):
        self.interactive_panel = InteractivePanel(self.model,
            viz=self.surface_visualizer_panel,
            ctviz=self.ct_visualizer_panel)

if __name__ == '__main__':
    #force Qt to relay ctrl+C
    from .main import main
    main()



#the intention of this file is to store operations on groups of electrodes here
#and call them from whatever place is appropriate

#i intend to move working code to this file on an as-tweaked basis

from .utils import ask_user_for_savefile
from .plotting_utils import coronal_slice
from functools import reduce

###############
# coronal_slice
###############

def coronal_slice_grid( electrodes, title=None, 
                        subjects_dir=None, subject=None, 
                        dpi=150, size=(200, 200), savefile=None):
    '''
    Helper function which calls the coronal_slice plotting function. Handles
    the case where the user specifies the geometry as any combination of 1 or
    2 dimensional matrices with only 1 meaningful vector.
    '''
    elecs_have_geom = True
    for elec in electrodes:
        if len(elec.geom_coords) != 2:
            elecs_have_geom = False
            break

    # we assume that we are only looking at depth leads with appropriate
    # 1xN geometry
    if elecs_have_geom:
        start = reduce( 
            lambda x,y:x if x.geom_coords[1]<y.geom_coords[1] else y,
            electrodes)
        end = reduce(
            lambda x,y:x if x.geom_coords[1]>y.geom_coords[1] else y,
            electrodes)
    else:
        start, end = (None, None)

    return coronal_slice( electrodes, start=start, end=end, outfile=savefile,
        subjects_dir=subjects_dir, subject=subject, title=title,
        dpi=dpi, size=size )

def coronal_slice_all( grids, grid_types, subjects_dir=None, subject=None,
                       dpi=150, size=(200, 200) ):
    '''
    takes a dictionary with grid assignments and creates a pdf with
    the coronal slices for these grids
    '''
    pdf_file = ask_user_for_savefile('save pdf file with coronal slice images')

    if pdf_file is None:
        return

    from matplotlib.backends.backend_pdf import PdfPages

    with PdfPages(pdf_file) as pdf:
        for grid in grids:

            if grid_types[grid] != 'depth':
                continue

            electrodes = grids[grid] 
            if len(electrodes) == 0:
                continue

            fig = coronal_slice_grid( electrodes, title=grid,
                subjects_dir=subjects_dir, subject=subject,
                dpi=dpi, size=size )

            pdf.savefig(fig)

#################
# get_nearby_rois
#################
def _find_surrounding_rois( elec, parcellation='aparc', error_radius=4,
                            subjects_dir=None, subject=None ):
    if elec.pial_coords is not None:
        pos = elec.pial_coords
    else:
        pos = elec.surf_coords

    from . import pipeline as pipe
    #TODO incorporate subcortical structures into non-aparc
    roi_hits = pipe.identify_roi_from_atlas( pos,
        atlas = parcellation,
        approx = error_radius,
        subjects_dir = subjects_dir,
        subject = subject)

    if len(roi_hits) == 0:
        roi_hits = ['None']

    elec.roi_list = roi_hits

def get_nearby_rois_elec( *args, **kwargs ):
    return _find_surrounding_rois(*args, **kwargs)

def get_nearby_rois_grid( electrodes, parcellation='aparc',
                          error_radius=4,
                          subjects_dir=None, subject=None ): 
    for elec in electrodes:
        try:
            _find_surrounding_rois( elec, 
                                    parcellation=parcellation,
                                    error_radius=error_radius,
                                    subjects_dir=subjects_dir,
                                    subject=subject )
        except:
            print('Failed to find ROIs for %s' % str(elec))

def get_nearby_rois_all( grids, subjects_dir=None, subject=None,
                         parcellation='aparc', error_radius=4 ):
    for grid in grids:    
        electrodes = grids[grid]
        get_nearby_rois_grid( electrodes,
                              parcellation=parcellation,
                              error_radius=error_radius,
                              subjects_dir=subjects_dir, subject=subject )

##############
# save_montage
##############

def _save_montage_file( savefile, electrodes, grid_types, 
                        snapping_completed=False ):
    # write the montage file
    with open( savefile, 'w' ) as fd:
        for j, elec in enumerate(electrodes):
            key = elec.grid_name
            if elec.name != '':
                label_name = elec.name
            else:
                elec_id = elec.geom_coords
                elec_2dcoord = ('unsorted%i'%j if len(elec_id)==0 else
                    str(elec_id))
                label_name = '%s_elec_%s'%(key, elec_2dcoord)

            if (snapping_completed and grid_types[key]=='subdural'):
                pos = elec.pial_coords.tolist()
            else:
                pos = tuple(elec.surf_coords)

            x,y,z = ['%.4f'%i for i in pos]

            line = '%s\t%s\t%s\t%s\n' % (label_name, x, y, z)

            fd.write(line)

def _save_csv_file( savefile, electrodes, grid_types,
                    snapping_completed=False ):
    #write the csv file
    import csv
    with open( savefile, 'w' ) as fd:
        writer = csv.writer(fd)

        for j,elec in enumerate(electrodes):
            key = elec.grid_name
            if elec.name != '':
                label_name = elec.name
            else:
                elec_id = elec.geom_coords
                elec_2dcoord = ('unsorted%i'%j if len(elec_id)==0 else
                    str(elec_id))
                label_name = '%s_elec_%s'%(key, elec_2dcoord)

            if (snapping_completed and grid_types[key]=='subdural'):
                pos = elec.pial_coords.tolist() 
            else:
                pos = tuple(elec.surf_coords)

            x,y,z = ['%.4f'%i for i in pos]

            row = [label_name, x, y, z]
            row.extend(elec.roi_list)

            writer.writerow(row)

def save_coordinates( electrodes, grid_types, snapping_completed=False,
                           file_type='csv'):
    savefile = ask_user_for_savefile(title='save %s file'%file_type)

    if savefile is None:
        return

    if file_type == 'csv':
        save_file_continuation = _save_csv_file
    elif file_type == 'montage':
        save_file_continuation = _save_montage_file

    save_file_continuation(savefile, electrodes, grid_types,
        snapping_completed=snapping_completed)

def save_coordinates_all( *args, **kwargs):
    return save_coordinates(*args, **kwargs)
def save_coordinates_grid( *args, **kwargs):
    return save_coordinates(*args, **kwargs)



import numpy as np
import os
from traits.api import (HasTraits, Str, Color, List, Instance, Int, Method,
    on_trait_change, Color, Any, Enum, Button, Float, File, Bool, Range,
    Event)
from traitsui.api import (View, Item, HGroup, Handler, CSVListEditor,
    InstanceEditor, Group, OKCancelButtons, TableEditor, ObjectColumn, 
    TextEditor, OKButton, CheckListEditor, OKCancelButtons, Label, Action,
    VSplit, HSplit, VGroup)
from traitsui.message import error as error_dialog

from functools import partial
from mayavi import mlab

class SortingLabelingError(Exception):
    pass

def virtual_points3d(coords, figure=None, scale_factor=None, color=None, 
    name=None):

    c = np.array(coords)
    source = mlab.pipeline.scalar_scatter( c[:,0], c[:,1], c[:,2],
        figure=figure)

    return mlab.pipeline.glyph( source, scale_mode='none', 
        scale_factor=scale_factor,
        mode='sphere', figure=figure, color=color, name=name)

    #return mlab.points3d( c[:,0], c[:,1], c[:,2],
    #    figure=figure, scale_factor=10*scale_factor, color=color, name=name)

def intize( tuple_of_floats ):
    return tuple( 
        map ( int,
        [x*1e4 for x in list(map( partial( round, ndigits=4 ), tuple_of_floats ))]))

def clear_scene(scene):
    #this bugs out for a reason I haven't figured out yet
    #for child in scene.children:
    #    child.remove() 

    #this causes a picker bug with or without gc
    #scene.children = []
    #import gc
    #gc.collect()

    #while scene.children: 
    #    for child in scene.children:
    #        child.remove()

    #Now we handle this by building a new viewport-like object
    #which CANNOT be done inside the viewport itself
    pass

    #mlab.clf(figure=scene)

def get_subjects_dir(subjects_dir=None, subject=None):
    if subjects_dir is None or subjects_dir=='':
        subjects_dir = os.environ['SUBJECTS_DIR']
    if subject is None or subject=='':
        subject = os.environ['SUBJECT']

    return os.path.join(subjects_dir, subject)

def _count():
    i=0
    while True:
        yield i
        i+=1
_counter=_count()
def gensym():
    global _counter
    return next(_counter)

def crash_if_freesurfer_is_not_sourced():
    import os, subprocess
    with open(os.devnull, 'w') as nil:
        p = subprocess.call(['which', 'mri_info'], stdout=nil)
    if p!=0:
        print('Freesurfer is not sourced or not in the subshell path')
        import sys

        print(os.environ['PATH'])

        sys.exit(1)

def ask_user_for_savefile(title=None):
    #from traitsui.file_dialog import save_file
    from pyface.api import FileDialog, OK
    
    dialog = FileDialog(action='save as')
    if title is not None:
        dialog.title = title
    dialog.open()
    if dialog.return_code != OK:
        return

    return os.path.join( dialog.directory, dialog.filename )

def ask_user_for_loadfile(title=None):
    from pyface.api import FileDialog, OK
    dialog = FileDialog(action='open')
    if title is not None:
        dialog.title = title
    dialog.open()
    if dialog.return_code != OK:
        return

    return os.path.join( dialog.directory, dialog.filename )

def get_default_color_scheme():
    predefined_colors = [(.2,.5,.8), #focal blue
                         (.6,.3,.9), #dark purple
                         (.8,.5,.9), #light purple
                         (1,.2,.5), #hot pink
                         (.7,.7,.9), #lavender
                         (.36,.58,.04), #dark green
                         (.22,.94,.64), #turquoise
                         (1,.6,.2), #orange
                         (.5,.9,.4), #semi-focal green
                         (0,.6,.8), #royal blue
                        ]

    for color in predefined_colors:
        yield color
    while True:
        yield tuple(np.random.random(3))

class AddLabelsWindow(Handler):
    model = Any
    #clumsily old a reference to the model object

    annotation = Str
    label = File

    add_annot_button = Button('Add annotation')
    add_label_button = Button('Add label file')

    annot_borders = Bool
    annot_opacity = Range(0., 1., 1.)
    annot_hemi = Enum('both','lh','rh')
    label_borders = Bool
    label_opacity = Range(0., 1., 1.)
    label_color = Color('blue')

    remove_labels_action = Action(name='Remove all labels', 
        action='do_remove')

    def _add_annot_button_fired(self):
        self.model.add_annotation(self.annotation, border=self.annot_borders,
            hemi=self.annot_hemi, opacity=self.annot_opacity)
         
    def _add_label_button_fired(self):
        self.model.add_label(self.label, border=self.label_borders,
            opacity=self.label_opacity, color=self.label_color)

    def do_remove(self, info):
        self.model.remove_labels()

    traits_view = View(
        HSplit(
        VGroup(
            Item('annotation'),
            Item('annot_borders', label='show border only'),
            Item('annot_opacity', label='opacity'),
            Item('annot_hemi', label='hemi'),
            Item('add_annot_button', show_label=False),
        ),
        VGroup(
            Item('label'),
            Item('label_borders', label='show_border_only'),
            Item('label_opacity', label='opacity'),
            Item('label_color', label='color'),
            Item('add_label_button', show_label=False),
        ),
        ),
        buttons=[remove_labels_action, OKButton],
        kind='livemodal',
        title='Dial 1-800-COLLECT and save a buck or two',
    )

#class RegistrationAdjustmentWindow(Handler):
#    model = Any
#    #we clumsily hold a reference to the model only to fire its events
#
#    cur_grid = Str
#
#    x_d = Button('-x')
#    x_i = Button('+x')
#    xval = Float(0.)
#    y_d = Button('-y')
#    y_i = Button('+y')
#    yval = Float(0.)
#    z_d = Button('-z')
#    z_i = Button('+z')
#    zval = Float(0.)
#
#    pitch_d = Button('pitch-')
#    pitch_i = Button('pitch+')
#    pitchval = Float(0.)
#    roll_d = Button('roll-')
#    roll_i = Button('roll+')
#    rollval = Float(0.)
#    yaw_d = Button('yaw-')
#    yaw_i = Button('yaw+')
#    yawval = Float(0.)
#
#    reset_button = Button('Reset')
#
#    cos5 = Float(np.cos(np.pi/36))
#    sin5 = Float(np.sin(np.pi/36))
#
#    traits_view = View(
#        Group(
#        HGroup(
#            Item('x_d'),
#            Item('xval'),
#            Item('x_i'),
#        show_labels=False),
#        HGroup(
#            Item('y_d'),
#            Item('yval'),
#            Item('y_i'),
#        show_labels=False),
#        HGroup(
#            Item('z_d'),
#            Item('zval'),
#            Item('z_i'),
#        show_labels=False),
#        HGroup(
#            Item('pitch_d'),
#            Item('pitchval'),
#            Item('pitch_i'),
#        show_labels=False),
#        HGroup(
#            Item('roll_d'),
#            Item('rollval'),
#            Item('roll_i'),
#        show_labels=False),
#        HGroup(
#            Item('yaw_d'),
#            Item('yawval'),
#            Item('yaw_i'),
#        show_labels=False),
#        Item('reset_button', show_label=False)
#        ), 
#    )
#
#    def _rot_matrix(self, axis, neg=False, deg=2):
#        cval = np.cos( deg*np.pi/180 )
#        sval = -np.sin( deg*np.pi/180 ) if neg else np.sin( deg*np.pi/180 )
#
#        if axis == 'x':
#            mat = np.array(( (1, 0, 0, 0),
#                             (0, cval, -sval, 0),
#                             (0, sval, cval, 0),
#                             (0, 0, 0, 1), ))
#        elif axis == 'y':
#            mat = np.array(( (cval, 0, sval, 0),
#                             (0, 1, 0, 0),
#                             (-sval, 0, cval, 0),
#                             (0, 0, 0, 1), ))
#        elif axis == 'z':
#            mat = np.array(( (cval, -sval, 0, 0),
#                             (sval, cval, 0, 0),
#                             (0, 0, 1, 0),
#                             (0, 0, 0, 1), ))
#        else:
#            return np.eye(4)
#
#        return mat
#
#    def _trans_matrix(self, axis, neg=False, dist=.5):
#        mat = np.eye(4)
#
#        if neg:
#            dist *= -1
#
#        if axis=='x': mat[0,3]=dist
#        elif axis=='y': mat[1,3]=dist
#        elif axis=='z': mat[2,3]=dist
#
#        return mat
#
#    def _x_d_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._trans_matrix('x', neg=True))
#    def _x_i_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._trans_matrix('x', neg=False))
#    def _y_d_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._trans_matrix('y', neg=True))
#    def _y_i_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._trans_matrix('y', neg=False))
#    def _z_d_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._trans_matrix('z', neg=True))
#    def _z_i_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._trans_matrix('z', neg=False))
#
#    def _roll_d_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._rot_matrix('x', neg=True) )
#    def _roll_i_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._rot_matrix('x', neg=False) )
#    def _pitch_d_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._rot_matrix('y', neg=True) )
#    def _pitch_i_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._rot_matrix('y', neg=False) )
#    def _yaw_d_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._rot_matrix('z', neg=True) )
#    def _yaw_i_fired(self):
#        self.model.reorient_glyph(self.cur_grid, 
#            self._rot_matrix('z', neg=False) )



import numpy as np
import os
from traits.api import (HasTraits, Str, Color, List, Instance, Int, Method,
    on_trait_change, Color, Any, Enum, Button, Float, File, Bool, Range,
    Event)
from traitsui.api import (View, Item, HGroup, Handler, CSVListEditor,
    InstanceEditor, Group, OKCancelButtons, TableEditor, ObjectColumn, 
    TextEditor, OKButton, CheckListEditor, OKCancelButtons, Label, Action,
    VSplit, HSplit, VGroup)
from traitsui.message import error as error_dialog

class NameHolder(HasTraits):
    name = Str
    traits_view = View()

    def __str__(self):
        return 'Grid: %s'%self.name

class GeometryNameHolder(NameHolder):
    geometry = Str
    color = Color
    previous_name = Str
    traits_view = View( 
        HGroup(
            Item('name', show_label=False, 
                editor=TextEditor(auto_set=False, enter_set=True), ),
            Item('geometry', style='readonly'),
            Item('color', style='readonly'),
        ),
    )

    def __str__(self):
        return 'Grid: %s, col:%s, geom:%s'%(self.name,self.color,
            self.geometry)
    def __repr__(self):
        return str(self)

class GeomGetterWindow(Handler):
    #has to do the handler thing
    #for now proof of concept and notimplementederror
    geometry = List(Int)
    holder = Instance(NameHolder)
    
    traits_view = View(
        #Item('grid_representation', editor=InstanceEditor(), style='custom'),
        Item('holder', editor=InstanceEditor(), style='custom', 
            show_label=False),
        Item('geometry', editor=CSVListEditor(), label='list geometry'),
        title='Specify geometry',
        kind='livemodal',
        buttons=OKCancelButtons,
    )

class NameHolderDisplayer(Handler):
    name_holders = List(Instance(NameHolder))
    interactive_mode = Instance(NameHolder)
    _mode_changed_event = Event

    @on_trait_change('interactive_mode')
    def fire_event(self):
        self._mode_changed_event = True

    traits_view = View(
        Item('interactive_mode', editor=InstanceEditor(name='name_holders'),
            style='custom', show_label=False),
    )




import numpy as np
from matplotlib.colors import LinearSegmentedColormap

def map_to_table(cmap,nvals=256):
    '''
    Takes a LinearSegmentedColormap, and returns a table of RGBA values
    spanning that colormap.

    arguments:
    cmap - the LinearSegmentedColormap instance
    nvals- the number of values to span over.  The default is 256.
    '''
    return cmap(range(nvals),bytes=True)

def mayavi2vtk_color(mayavi_color):
    '''
    converts a 3-tuple mayavi color with values [0,1] to a 4-tuple RGBA
    VTK color with float values [0.0, 255.0]
    '''
    rgb_col = [round(255*color) for color in mayavi_color]
    rgba_col = (rgb_col[0], rgb_col[1], rgb_col[2], 255.0)
    return rgba_col

def mayavi2traits_color(mayavi_color):
    '''
    converts a 3-tuple mayavi color with values [0,1] to a 3-tuple RGB
    color with integer values [0, 255]

    does not use QColor objects at all currently
    '''
    return tuple([int(255*color) for color in mayavi_color])

def traits2mayavi_color(traits_color):
    '''
    Queries the backend:
        wx:
    converts a 4-tuple traits color with integer values [0, 255] to a
    3-tuple mayavi color with values [0,1]
        Qt:
    converts a 4-tuple traits color with integer values [0, 255] embedded
    inside a QColor object to a 3-tuple mayavi color with values [0,1]
    '''
    from traits.trait_base import ETSConfig
    _tk = ETSConfig.toolkit
    if _tk == 'wx':
        rgba_col = [color/255 for color in traits_color]
    elif _tk == 'qt4':
        rgba_col = [color/255 for color in traits_color.getRgb()]
    rgb_col = (rgba_col[0], rgba_col[1], rgba_col[2])
    return rgb_col 

def set_discrete_lut(mayavi_obj, colors, use_vector_lut=False):
    '''
    sets the lookup table of the mayavi object provided to an N color
    color map with the provided list of colors. Also sets the data range
    trait of the LUT.

    Parameters
    ----------
    mayavi_obj : Instance(mlab.pipeline)
        A mayavi object with a module_manager instance
    colors : List(3-tuple)
        List of RGB colors as 3-tuples with values [0,1].
        The color at index 0 corresponds to scalar value 0, and so on
            up to scalar value N-1
    use_vector_lut : Bool
        If true, uses the vector_lut_manager. Otherwise, uses the
        scalar_lut_manager. Defaults to False.
    '''
    if use_vector_lut:
        lut_mgr=mayavi_obj.module_manager.vector_lut_manager
    else:
        lut_mgr=mayavi_obj.module_manager.scalar_lut_manager

    mayavi_obj.actor.mapper.scalar_visibility = True

    n = len(colors)
    #cmap = LinearSegmentedColormap.from_list('ign', colors)
    cmap = list(map(mayavi2vtk_color, colors))

    #lut_mgr.lut.table = map_to_table(cmap, nvals=n)
    lut_mgr.lut.table = cmap
    lut_mgr.number_of_colors = n
    lut_mgr.data_range = [0, n-1]

def make_transparent(mayavi_obj, index, use_vector_lut=False):
    '''
    Make transparent the color at specified index in the provided LUT.
    Otherwise expects a well formed LUT.
    '''
    if use_vector_lut:
        lut_mgr=mayavi_obj.module_manager.vector_lut_manager
    else:
        lut_mgr=mayavi_obj.module_manager.scalar_lut_manager

    mayavi_obj.actor.mapper.scalar_visibility = True
    
    table = lut_mgr.lut.table
    table[index] = (0., 0., 0., 0.)
    lut_mgr.lut.table = np.array(table).tolist()

    #why in the hell is this necessary
    #without these lines, the hiding occurs on the first time, but on no
    #subsequent times. it's not a race condition either because the random
    #arithmetic below doesn't result in success.
    n = len(table)
    lut_mgr.number_of_colors = n
    lut_mgr.data_range = [0, n-1]

    #f = 37
    #g = f**16 + 21

def change_single_glyph_color(mayavi_glyph, index, color):
    '''
    changes the scalars trait of a mayavi glyph safely to display the altered
    color of a single node

    Parameters
    ----------

    mayavi_glyph : mayavi.modules.Glyph
        The glyph object
    index : int
        The offset of the node within the glyph
    color : number
        The new value of the scalar to set at this index
    '''
    colors = np.array(mayavi_glyph.mlab_source.dataset.point_data.scalars)
    colors[index] = color
    mayavi_glyph.mlab_source.dataset.point_data.scalars = colors

def set_binary_lut(mayavi_obj, color1, color2, use_vector_lut=False):
    '''
    sets the lookup table to of the mayavi object provided to a 2 color
    color map ranging between color1 and color2. Also sets the appropriate
    attributes of the LUT so that the scalars are immediately displayed

    Parameters
    ----------
    mayavi_obj : Instance(mlab.pipeline)
        A mayavi object with a module_manager instance
    color1 : 3-tuple
        An RGB color represented as a 3-tuple with values [0,1], and
        corresponding to the scalar value 0
    color2 : 3-tuple
        An RGB color represented as a 3-tuple with values [0,1], and
        corresponding to the scalar value 1
    use_vector_lut : Bool
        If true, uses the vector_lut_manager. Otherwise, uses the
        scalar_lut_manager. Defaults to False.
    '''
    mayavi_obj.actor.mapper.scalar_visibility = True

    if use_vector_lut:
        lut_mgr=mayavi_obj.module_manager.vector_lut_manager
    else:
        lut_mgr=mayavi_obj.module_manager.scalar_lut_manager

    lut_mgr.lut.table = map_to_table(
        LinearSegmentedColormap.from_list('ign', [color1, color2]))
        #map(mayavi2vtk_color, [color1, color2])))

    #lut_mgr.use_default_range = False
    lut_mgr.data_range = [0,1]

def set_monochrome_lut(mayavi_obj):
    '''
    sets the attributes of the mayavi object so that scalar mapped
    colors are not displayed
    '''
    mayavi_obj.actor.mapper.scalar_visibility = False



import os
from traits.trait_base import ETSConfig
_tk = ETSConfig.toolkit

from functools import partial
from traits.api import (Property, Instance, Bool, Str, BaseTraitHandler,
    Range)
from traitsui.api import CustomEditor

if _tk in ('null','',None):
    _tk = os.environ['ETS_TOOLKIT']
    #there's no situation other than debugging this file directly
    #where this won't be set. To debug this file, set the environment var.
else:
    print(_tk)

CustomEditorKlass = __import__('traitsui.%s.custom_editor'%_tk,
    fromlist=['CustomEditor'], ).CustomEditor

#def mkeditor(*args, **kwargs):
#    if _tk == 'qt4':
#        klass_factory_fun = qt4_editor_factory
#    else:
#        klass_factory_fun = wx_editor_factory
#
#    return klass_factory_fun(*args, **kwargs)

def mkeditor(*args, **kwargs):
    custom_editor_replacement_klass = __import__('ielu.custom_list_editor',
        fromlist=['ielu'])

    klass_factory_fun = getattr(custom_editor_replacement_klass, 
        '{0}_editor_factory'.format(_tk))

    return klass_factory_fun(*args, **kwargs)

#    return getattr(__import__('custom_list_editor', fromlist=['ielu']),
#        '{0}_editor.factory'.format(_tk))(*args, **kwargs)

def wx_editor_factory(*args, **kwargs):
    pass

def qt4_editor_factory(parent, editor, *args, **kwargs):
    from pyface.qt import QtCore, QtGui
    trait_handler = editor.factory.trait_handler
    if trait_handler is None:
        trait_handler = editor.object.base_trait( editor.name ).handler
    editor._trait_handler = trait_handler

    editor.control = panel = QtGui.QScrollArea()
    editor.control.setFrameShape(QtGui.QFrame.NoFrame)
    editor.control.setWidgetResizable(True)

    editor.mapper = QtCore.QSignalMapper(panel)

    editor._list_pane = QtGui.QWidget()
    editor._list_pane.setSizePolicy(QtGui.QSizePolicy.Expanding,
                                  QtGui.QSizePolicy.Expanding)
    layout = QtGui.QGridLayout(editor._list_pane)
    layout.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop) 
    layout.setContentsMargins(0, 0, 0, 0)

    _editor = editor.factory.editor
    if _editor is None:
        _editor = trait_handler.item_trait.get_editor()
    editor._editor = getattr( _editor, editor.kind )

    extended_name = editor.extended_name.replace('.', ':')
    editor.context_object.on_trait_change( editor.update_editor_item,
                           extended_name + '_items?', dispatch = 'ui' )
    editor.set_tooltip()

    return panel

#custom editor on toolkit level
class CustomQtListEditorKlass(CustomEditorKlass):

    from pyface.qt import QtCore, QtGui

    #trait definitions copied from file
    mapper = Instance(QtCore.QSignalMapper)
    mutable = Bool(True)
    kind = Str
    
    single_row = True

    list_menu = """
       Add &Before     [_menu_before]: self.add_before()
       Add &After      [_menu_after]:  self.add_after()
       ---
       &Delete         [_menu_delete]: self.delete_item()
       ---
       Move &Up        [_menu_up]:     self.move_up()
       Move &Down      [_menu_down]:   self.move_down()
       Move to &Top    [_menu_top]:    self.move_top()
       Move to &Bottom [_menu_bottom]: self.move_bottom()
    """

    empty_list_menu = """
       Add: self.add_empty()
    """

    def update_editor(self): 
        from pyface.qt import QtCore, QtGui
        self.mapper = QtCore.QSignalMapper(self.control)
        # Disconnect the editor from any control about to be destroyed:
        self._dispose_items()

        list_pane = self._list_pane
        layout = list_pane.layout()

        # Create all of the list item trait editors:
        trait_handler = self._trait_handler
        resizable     = ((trait_handler.minlen != trait_handler.maxlen) and
                         self.mutable)
        item_trait    = trait_handler.item_trait

        is_fake = (resizable and (len( self.value ) == 0))
        if is_fake:
            self.empty_list()
        else:
            # Asking the mapper to send the sender to the callback method
            self.mapper.mapped[QtCore.QObject].connect(self.popup_menu)

        editor = self._editor
        for index, value in enumerate(self.value):
            row, column = divmod(index, self.factory.columns)

            # Account for the fact that we have <columns> number of
            # pairs
            column = column * 2

            if resizable:
                # Connecting the new button to the mapper
                from traitsui.qt4.helper import IconButton

                from pyface.qt import qt_api

                # fix bug in pyqt, reverts back to traitsui 4.3 version
                if qt_api == 'pyside':
                    control = IconButton('list_editor.png', self.mapper.map)
                elif qt_api in ('pyqt', 'pyqt5'):
                    control = IconButton('list_editor.png',
                        lambda : self.popup_menu(
                            self._list_pane.layout().sender() ))

                # Setting the mapping and asking it to send the index of the
                # sender to the callback method
                self.mapper.setMapping(control, control)

                layout.addWidget(control, row, column)

            from traitsui.editors.list_editor import ListItemProxy
            proxy = ListItemProxy( self.object, self.name, index, item_trait,
                                   value )
            if resizable:
                control.proxy = proxy
            peditor = editor( self.ui, proxy, 'value', self.description,
                              list_pane ).set( object_name = '' )
            peditor.prepare( list_pane )
            pcontrol = peditor.control
            pcontrol.proxy = proxy

            if isinstance(pcontrol, QtGui.QWidget):
                layout.addWidget(pcontrol, row, column+1)
            else:
                layout.addLayout(pcontrol, row, column+1)

        # QScrollArea can have problems if the widget being scrolled is set too
        # early (ie. before it contains something).
        if self.control.widget() is None:
            self.control.setWidget(list_pane)

    def update_editor_item ( self, event ):
        """ Updates the editor when an item in the object trait changes
        externally to the editor.
        """
        # If this is not a simple, single item update, rebuild entire editor:
        from pyface.qt import QtCore, QtGui
        if (len( event.removed ) != 1) or (len( event.added ) != 1):
            self.update_editor()
            return

        # Otherwise, find the proxy for this index and update it with the
        # changed value:
        for control in self.control.widget().children():
            if isinstance(control, QtGui.QLayout):
                continue

            proxy = control.proxy
            if proxy.index == event.index:
                proxy.value = event.added[0]
                break

    def dispose ( self ):
        """ Disposes of the contents of an editor.
        """
        self._dispose_items()

        extended_name = self.extended_name.replace('.', ':')
        self.context_object.on_trait_change( self.update_editor_item,
                                 extended_name + '_items?', remove = True )

        super( CustomQtListEditorKlass, self ).dispose()

    def empty_list ( self ):
        """ Creates an empty list entry (so the user can add a new item).
        """
        # Connecting the new button to the mapper
        from pyface.qt import QtCore, QtGui
        from pyface.qt import qt_api
        from traitsui.qt4.helper import IconButton

        #control = IconButton('list_editor.png', self.mapper.map)
        if qt_api == 'pyside':
            control = IconButton('list_editor.png', self.mapper.map)
        elif qt_api == 'pyqt':
            control = IconButton('list_editor.png',
                lambda : self.popup_empty_menu(
                    self._list_pane.layout().sender() ))

        # Setting the mapping and asking it to send the sender to the
        # callback method
        self.mapper.setMapping(control, control)
        self.mapper.mapped[QtCore.QObject].connect(self.popup_empty_menu)
        control.is_empty = True
        self._cur_control = control

        from traitsui.editors.list_editor import ListItemProxy
        proxy    = ListItemProxy( self.object, self.name, -1, None, None )
        pcontrol = QtGui.QLabel('   (Empty List)')
        pcontrol.proxy = control.proxy = proxy

        layout = self._list_pane.layout()
        layout.addWidget(control, 0, 0)
        layout.addWidget(pcontrol, 0, 1)

    def get_info ( self ):
        """ Returns the associated object list and current item index.
        """
        proxy = self._cur_control.proxy
        return ( proxy.list, proxy.index )

    def popup_empty_menu ( self , sender):
        """ Displays the empty list editor popup menu.
        """
        from traitsui.qt4.menu import MakeMenu
        from pyface.qt import QtCore, QtGui
        self._cur_control = control = sender
        menu = MakeMenu( self.empty_list_menu, self, True, control ).menu
        menu.exec_(control.mapToGlobal(QtCore.QPoint(0, 0)))

    def popup_menu ( self , sender):
        """ Displays the list editor popup menu.
        """
        from traitsui.qt4.menu import MakeMenu
        from pyface.qt import QtCore, QtGui
        self._cur_control = sender

        proxy    = sender.proxy
        index    = proxy.index
        menu     = MakeMenu( self.list_menu, self, True, sender ).menu
        len_list = len( proxy.list )
        not_full = (len_list < self._trait_handler.maxlen)

        self._menu_before.enabled( not_full )
        self._menu_after.enabled(  not_full )
        self._menu_delete.enabled( len_list > self._trait_handler.minlen )
        self._menu_up.enabled(  index > 0 )
        self._menu_top.enabled( index > 0 )
        self._menu_down.enabled(   index < (len_list - 1) )
        self._menu_bottom.enabled( index < (len_list - 1) )

        menu.exec_(sender.mapToGlobal(QtCore.QPoint(0, 0)))

    def add_item ( self, offset ):
        """ Adds a new value at the specified list index.
        """
        list, index = self.get_info()
        index      += offset
        item_trait  = self._trait_handler.item_trait
        value       = item_trait.default_value_for( self.object, self.name )
        self.value = list[:index] + [ value ] + list[index:]
        self.update_editor()

    def add_before ( self ):
        """ Inserts a new item before the current item.
        """
        self.add_item( 0 )

    def add_after ( self ):
        """ Inserts a new item after the current item.
        """
        self.add_item( 1 )

    def add_empty ( self ):
        """ Adds a new item when the list is empty.
        """
        list, index = self.get_info()
        self.add_item( 0 )

    def delete_item ( self ):
        """ Delete the current item.
        """
        list, index = self.get_info()
        self.value  = list[:index] + list[index+1:]
        self.update_editor()

    def move_up ( self ):
        """ Move the current item up one in the list.
        """
        list, index = self.get_info()
        self.value  = (list[:index-1] + [ list[index], list[index-1] ] +
                       list[index+1:])
        self.update_editor()

    def move_down ( self ):
        """ Moves the current item down one in the list.
        """
        list, index = self.get_info()
        self.value  = (list[:index] + [ list[index+1], list[index] ] +
                       list[index+2:])
        self.update_editor()

    def move_top ( self ):
        """ Moves the current item to the top of the list.
        """
        list, index = self.get_info()
        self.value  = [ list[index] ] + list[:index] + list[index+1:]
        self.update_editor()

    def move_bottom ( self ):
        """ Moves the current item to the bottom of the list.
        """
        list, index = self.get_info()
        self.value  = list[:index] + list[index+1:] + [ list[index] ]
        self.update_editor()

    def _dispose_items ( self ):
        """ Disposes of each current list item.
        """
        layout = self._list_pane.layout()
        child = layout.takeAt(0)
        while child is not None:
            control = child.widget()
            if control is not None:
                editor = getattr( control, '_editor', None )
                if editor is not None:
                    editor.dispose()
                    editor.control = None
                control.deleteLater()
            child = layout.takeAt(0)
        del child

    def _kind_default(self):
        """ Returns a default value for the 'kind' trait.
        """
        return self.factory.style + '_editor'

    def _mutable_default(self):
        """ Trait handler to set the mutable trait from the factory.
        """
        return self.factory.mutable

if _tk == 'wx':
    from traitsui.editors.list_editor import ListEditor as CustomListEditor
elif _tk == 'qt4':

    from traitsui.editor_factory import EditorFactory
    from traitsui.ui_traits import style_trait

    #custom edior on traitsui level
    class CustomListEditor(CustomEditor):
        '''
        Traitsui level abstraction layer for the Custom List Editor
        '''
        factory = Property #Callable

        editor = Instance( EditorFactory )

        mutable = Bool(True)

        style = style_trait

        trait_handler = Instance( BaseTraitHandler )

        rows = Range(1, 50, 5, desc = 'the number of list rows to display')
        columns = Range(1, 10, 1, dec = 'the number of list cols to display')

        use_notebook = Bool(False)
        show_notebook_menu = Bool(False)

        def _get_klass(self):
            #tell the editor to use the klass defined in this file
            return CustomQtListEditorKlass

        def _get_factory(self):
            return partial(mkeditor)
        
if __name__=='__main__':
    import signal
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    from traits.api import List, HasTraits
    from traitsui.api import View, Item 

    class Dog(HasTraits):
        dog = List

        def _dog_default(self):
            return [4,5,7]

        view = View(Item('dog', editor=CustomListEditor()))

    Dog().configure_traits()



__version__ = '0.3.0'

from .gselu import ElectrodePositionsModel

#import gselu
#from gselu import ElectrodePositionsModel
#
#import sys
#sys.modules['gselu'] = gselu
#sys.modules['gselu.gselu'] = gselu
#sys.modules['gselu.ElectrodePositionsModel'] = ElectrodePositionsModel
#sys.modules['gselu.gselu.ElectrodePositionsModel'] = ElectrodePositionsModel
#sys.modules['ElectrodePositionsModel'] = ElectrodePositionsModel



#!/usr/bin/env python



from xvfbwrapper import Xvfb

vdisplay = Xvfb()
vdisplay.start()

import os
import sys
from .gselu import ElectrodePositionsModel
from . import pipeline as pipe

ct_dir = '/space/truffles/2/users/rlaplant/ct'


#1. read batch file, consisting of subject (pickle) and scope of parameters

model = ElectrodePositionsModel()
if len(sys.argv) != 3:
    raise RuntimeError('provide batch file and batch symbol as argument')

batch_file = sys.argv[1]

with open(batch_file) as fd:
    batch_data = fd.readlines()

subject, delta, rho, tau, epsilon = [x.strip() for x in batch_data]

batch_symbol = sys.argv[2]

pickles_dir = os.path.join(ct_dir, 'batch_files')
#batch_dir = os.path.join(pickles_dir, '{0}_batch'.format(batch_symbol))
output_dir = os.path.join(ct_dir, 'output_files', 
    '{0}_output'.format(batch_symbol))

pickle_file = os.path.join(pickles_dir, '{0}_batch.pkl'.format(subject))
delta = float(delta)
rho = float(rho)
tau = float(tau) / 100
epsilon = float(epsilon)

if rho < 15:
    rho_strict = 5
    rho_loose = rho + 15
elif rho > 75:
    rho_strict = rho - 15
    rho_loose = 85
else:
    rho_strict = rho - 15
    rho_loose = rho + 15
    

#2. load pickle data

from pickle import load

with open(pickle_file) as fd:
    model = load(fd)

true_grids = model._grids
true_grid_geom = [x for x in list(model._grid_geom.values()) if x!='user-defined']

print(true_grid_geom)

all_electrodes = list(model._all_electrodes.values())

#3. run grid algorithm with specified data and parameters

print(delta, rho, tau, epsilon)

try:
    colors, grid_geom, new_grids, color_scheme = pipe.classify_electrodes(  all_electrodes,
                                                true_grid_geom,
                                                delta = delta,
                                                epsilon = epsilon,
                                                rho = rho,
                                                crit_pct = tau,
                                                rho_strict = rho_strict,
                                                rho_loose = rho_loose )
    #if anything goes bad then set 0
except:
    new_grids = {}

#from PyQt4.QtCore import pyqtRemoveInputHook
#pyqtRemoveInputHook()
#import pdb
#pdb.set_trace()

#4. calculate correctness

electrode_scores = []

for true_grid in list(true_grids.values()):
    for elec in true_grid:

        grid_concordance = 0
        grid_size = len(true_grid)

        #find provisional grid

        provisional_grid = None
        #for new grid in new_grids:
        for new_grid in list(new_grids.values()):
            if elec in new_grid:
                provisional_grid = new_grid
                break

        #if we failed to find the electrode in the true solution this
        #electrode is assigned a score of 0
        if provisional_grid is None:
            score = 0
            electrode_scores.append(score)
            break 

        # calculate completeness
        for reference_elec in true_grid:
            if reference_elec in provisional_grid:
                grid_concordance += 1
        
        score = grid_concordance / grid_size
        electrode_scores.append(score)

nr_electrodes = len(electrode_scores)
solution_score = sum(electrode_scores) / nr_electrodes

#5. save results to disk

output_file = os.path.join(output_dir, '{0}_{1}_{2}_{3}_{4}.output'.format(
                    subject,
                    delta,
                    rho,
                    tau,
                    epsilon))

with open(output_file, 'w') as fd:
    fd.write(str(solution_score))


if len(new_grids) > 0:

    out_pickle_file = os.path.join(output_dir, 
        '{0}_optimized_output.pkl'.format(subject))

    with open(out_pickle_file, 'w') as fd:
        model._colors = colors
        model._grid_geom = grid_geom
        model._grids = new_grids
        model._color_scheme = color_scheme

        pickle.dump(model, fd)



vdisplay.stop()



from traitsui.api import toolkit
from traits.trait_base import ETSConfig

from .gselu import iEEGCoregistrationFrame
from .utils import crash_if_freesurfer_is_not_sourced
import signal


def main():
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    crash_if_freesurfer_is_not_sourced()

    iEEGCoregistrationFrame().configure_traits()




import os
import numpy as np
import nibabel as nib

from traits.api import HasTraits, Float, Int, Tuple
from traitsui.api import View, Item, CSVListEditor

from .geometry import get_vox2rasxfm, apply_affine, get_std_orientation
from .utils import get_subjects_dir

def force_render( figure=None ):
    from mayavi import mlab
    figure.scene.render()
    mlab.draw(figure=figure)
    from pyface.api import GUI
    _gui = GUI()
    orig_val = _gui.busy
    _gui.set_busy(busy=True)
    _gui.process_events()
    _gui.set_busy(busy=orig_val)
    _gui.process_events()

def coronal_slice(elecs, start=None, end=None, outfile=None, 
    subjects_dir=None,
    subject=None, reorient2std=True, dpi=150, size=(200,200),
    title=None): 
    '''
    create an image of a coronal slice which serves as a guesstimate of a
    depth lead inserted laterally and nonvaryingly in the Y axis

    plot the electrodes from the lead overlaid on the slice in the X and Z
    directions

    Paramaters
    ----------
    elecs : List( Electrode )
        list of electrode objects forming this depth lead
    start : Electrode
        Electrode object at one end of the depth lead
    end : Electrode
        Electrode object at the other end of the depth lead
    outfile : Str
        Filename to save the image to
    subjects_dir : Str | None
        The freesurfer subjects_dir. If this is None, it is assumed to be the
        $SUBJECTS_DIR environment variable. If this folder is not writable,
        the program will crash.
    subject : Str | None
        The freesurfer subject. If this is None, it is assumed to be the
        $SUBJECT environment variable.
    reorient2std : Bool
        Apply a matrix to rotate orig.mgz to the standard MNI orientation
        emulating fslreorient2std. Pretty much always true here.
    dpi : Int
        Dots per inch of output image
    size : Tuple
        Specify a 2-tuple to control the image size, default is (200,200)
    title : Str
        Specify a matplotlib title
    '''
    print('creating coronal slice with start electrodes %s' % str(start))

    subjdir_subj = get_subjects_dir( subjects_dir=subjects_dir,
                                     subject=subject )
    orig = os.path.join(subjdir_subj, 'mri', 'orig.mgz')

    x_size, y_size, z_size = nib.load(orig).shape

    # vox2ras and ras2vox shouldnt have different procedures for 
    # getting the different dimensions. the matrix showing those
    # dimensions has the correct dimensions by inversion beforehand
    # in the complex 3-way case
    vox2ras = get_vox2rasxfm(orig, stem='vox2ras')
    ras2vox = np.linalg.inv(vox2ras)

    ras2vox[0:3,3] = (x_size/2, y_size/2, z_size/2)

    rd, ad, sd = get_std_orientation(ras2vox)

#    rd, = np.where(np.abs(ras2vox[:,0]) == np.max(np.abs(ras2vox[:,0])))
#    ad, = np.where(np.abs(ras2vox[:,1]) == np.max(np.abs(ras2vox[:,1])))
#    sd, = np.where(np.abs(ras2vox[:,2]) == np.max(np.abs(ras2vox[:,2])))

    r_size = [x_size, y_size, z_size][rd]
    a_size = [x_size, y_size, z_size][ad]
    s_size = [x_size, y_size, z_size][sd]

    #starty = pd.map_cursor( start.asras(), pd.current_affine, invert=True)[1]
    #endy = pd.map_cursor( end.asras(), pd.current_affine, invert=True )[1]
    #midy = (starty+endy)/2
    #pd.move_cursor(128, midy, 128)

    electrodes = np.squeeze([apply_affine([e.asras()], ras2vox) 
        for e in elecs])
    #electrodes = np.array([pd.map_cursor(e.asras(), ras2vox,
    #    invert=True) for e in elecs])

    vol = np.transpose( nib.load(orig).get_data(), (rd, ad, sd) )
    
    if start is not None and end is not None:
        start_coord = np.squeeze(apply_affine([start.asras()], ras2vox))
        end_coord = np.squeeze(apply_affine([end.asras()], ras2vox))

        if start_coord[rd] == end_coord[rd]:
            raise ValueError('This lead has no variation in the X axis. It shouldnt be displayed coronally')
        
        slice = np.zeros((s_size, r_size))
        
        m = (start_coord[ad]-end_coord[ad])/(start_coord[rd]-end_coord[rd])
        b = start_coord[ad]-m*start_coord[rd]

        rnew = np.arange(r_size)
        anew = m*rnew+b
        alower = np.floor(anew)
        afrac = np.mod(anew, 1)

        try:
            for rvox in rnew:
                slice[:, rvox] = (vol[rvox, alower[rvox], :] * 
                    (1-afrac[rvox])+vol[rvox, alower[rvox]+1, :] *
                    afrac[rvox])
        except IndexError:
            raise ValueError('This lead has minimal variation in the X axis. It shouldnt be displayed coronally')

    else:
        slice_nr = np.mean(electrodes[:,ad])
        slice = vol[:, slice_nr, :].T
    
    vox2pix = np.zeros((2,4))
    vox2pix[0, rd] = 1
    vox2pix[1, sd] = 1
    ras2pix = np.dot(vox2pix, ras2vox)

    pix = np.dot(ras2pix, 
        np.transpose([np.append(e.asras(), 1) for e in elecs]))

    #add data to coronal plane
    import pylab as pl

    fig = pl.figure()

    pl.imshow(slice, cmap='gray')
    pl.scatter(pix[0,:], pix[1,:], s=10, c='red', edgecolor='yellow',
        linewidths=0.4)

    if title is not None:
        pl.title(title)

    pl.axis('off')
    #pl.show()

    if outfile is not None:
        pl.savefig(outfile, dpi=dpi)

    return fig

def sequence_3d_images( figure ):
    from mayavi import mlab
    views = [lambda:mlab.view( azimuth=0, elevation=90, figure=figure ),
        lambda:mlab.view( azimuth=180, elevation=90, figure=figure ),
        lambda:mlab.view( azimuth=0, elevation=0, figure=figure ),
        lambda:mlab.view( azimuth=90, elevation=90, figure=figure ),
        lambda:mlab.view( azimuth=270, elevation=90, figure=figure )]

    for view in views:
        yield view

def save_opaque_clinical_sequence( savefile, mayavi_figure ):
    import pylab as pl
    from matplotlib.backends.backend_pdf import PdfPages
    from mayavi import mlab

    with PdfPages(savefile) as pdf:
        for angle in sequence_3d_images( mayavi_figure ):
            angle()
            force_render( figure=mayavi_figure )
            pixmap = mlab.screenshot( figure=mayavi_figure )
            mpl_figure = pl.figure()
            pl.imshow(pixmap, figure=mpl_figure)
            pdf.savefig(mpl_figure)




#!/usr/bin/env python
import sys
import numpy as np
import nibabel as nib
from scipy.signal import convolve
from scipy.ndimage.morphology import grey_closing
from scipy.ndimage.morphology import generate_binary_structure
from mne import write_surface
from mcubes import marching_cubes

def mkoutersurf(image, radius, outfile):
    #radius information is currently ignored
    #it is a little tougher to deal with the morphology in python

    fill = nib.load( image )
    filld = fill.get_data()
    filld[filld==1] = 255

    gaussian = np.ones((2,2))*.25

    image_f = np.zeros((256,256,256))

    for slice in range(256):
        temp = filld[:,:,slice]
        image_f[:,:,slice] = convolve(temp, gaussian, 'same')

    image2 = np.zeros((256,256,256))
    image2[np.where(image_f <= 25)] = 0
    image2[np.where(image_f > 25)] = 255

    strel15 = generate_binary_structure(3, 1)

    BW2 = grey_closing(image2, structure=strel15)
    thresh = np.max(BW2)/2
    BW2[np.where(BW2 <= thresh)] = 0
    BW2[np.where(BW2 > thresh)] = 255

    v, f = marching_cubes(BW2, 100)

    v2 = np.transpose(
             np.vstack( ( 128 - v[:,0],
                          v[:,2] - 128,
                          128 - v[:,1], )))
    
    write_surface(outfile, v2, f)

if __name__=='__main__':

    if not len(sys.argv) == 4:
        raise ValueError("the fail at give argument correct,\n"
            "volume filled first, diameter integral then, arguments file "
            "output")

    mkoutersurf( sys.argv[1], None, sys.argv[3] )

    




import numpy as np
from traits.api import (HasTraits, List, Float, Tuple, Instance, Bool, Str, 
    Int, Either, Property, Method, on_trait_change, Any, Enum, Button)
from traitsui.api import (View, Item, HGroup, Handler, CSVListEditor, VGroup,
    InstanceEditor, Group, OKCancelButtons, TableEditor, ObjectColumn,
    TextEditor, OKButton, CheckListEditor, Label, Action, ListStrEditor,
    MenuBar, Menu)
from traitsui.message import error as error_dialog
from .utils import ask_user_for_savefile
from functools import partial, total_ordering

class Electrode(HasTraits):
#    ct_coords = List(Float)
#    surf_coords = List(Float)
#    snap_coords = List(Float)
    ct_coords = Tuple
    surf_coords = Tuple

    iso_coords = Tuple

    #snap coords are an intermediate field used during snapping
    #the snap coords are on the dural surface and may also be more useful
    #for visualization
    #pial coords are where the final result is stored

    snap_coords = Either(None, Instance(np.ndarray))
    pial_coords = Either(None, Instance(np.ndarray))

    special_name = Str

    is_interpolation = Bool(False)
    grid_name = Str('unsorted')
    grid_transition_to = Str('')
    
    hemi = Str
    vertno = Int(-1)

    plane_coords = Either(None, Tuple)
    #geom_coords = Either(None, Tuple)
    geom_coords = List(Int)

    name = Str
    corner = List

    roi_list = List(Str)

    default_repr = Property
    def _get_default_repr(self):
        if self.special_name != '':
            return self.special_name
        return self._coord_repr('ct_coords')

    def _coord_print(self, coord_type):
        return ', '.join(
            ['{0:.2f}'.format(c) for c in getattr(self, coord_type)])

    def _coord_repr(self, coord_type):
        return 'Elec: {0} ({1})'.format(
            self.grid_name,
            self._coord_print(coord_type))

    def __str__(self):
        return self._coord_repr('ct_coords')
    
    def __repr__(self):
        return self.__str__()

    ras_repr = Property
    def _get_ras_repr(self):
        return self._coord_print('surf_coords')

    iso_repr = Property
    def _get_iso_repr(self):
        return self._coord_print('iso_coords')

    postsnap_repr = Property
    def _get_postsnap_repr(self):
        return self._coord_print('pial_coords')

    def __eq__(self, other):
        if other is None and self is None:
            return True
        if None in (other, self):
            return False
        if self.name != '' and other.name != '':
            return self.name == other.name
        else:
            return str(self) == str(other)

    def __lt__(self, other):
        if None in (other, self):
            return False
        if self.name != '' and other.name != '':
            return self.name < other.name
        else:
            return str(self) < str(other)

    def __gt__(self, other):
        if None in (other, self):
            return False
        if self.name != '' and other.name != '':
            return self.name > other.name
        else:
            return str(self) > str(other)
   
    def __le__(self, other):
        if None in (other, self):
            return False
        if self.name != '' and other.name != '':
            return self.name <= other.name
        else:
            return str(self) <= str(other)
    
    def __ge__(self, other):
        if None in (other, self):
            return False
        if self.name != '' and other.name != '':
            return self.name >= other.name
        else:
            return str(self) >= str(other)

    def astuple(self):
        #snap_coords is an intermediate variable. the final value of
        #the snapped electrodes on the pial surface is held in pial coords
        return nparrayastuple(self.pial_coords)

    def asras(self):
        return self.surf_coords

    def asct(self):
        return self.ct_coords

    def asiso(self):
        return self.iso_coords

def nparrayastuple(nparray):
    nparray = np.array(nparray)
    return (nparray[0], nparray[1], nparray[2])

class ElectrodeWindow(Handler):
    model = Any
    #we clumsily hold a reference to the model object to fire its events

    cur_grid = Str

    electrodes = List(Instance(Electrode))
    cur_sel = Instance(Electrode)
    selection_callback = Method
   
    selected_ixes = Any
    swap_action = Action(name='Swap two electrodes', action='do_swap')
    add_blank_action = Action(name='Add blank electrode', 
        action='do_add_blank')

    previous_sel = Instance(Electrode)
    previous_color = Int

    distinct_prev_sel = Instance(Electrode)
    
    save_montage_action = Action(name='Save montage file', 
        action='do_montage')

    save_csv_action = Action(name='Save CSV file', action='do_csv')

    interpolate_action = Action(name='Linear interpolation',
        action='do_linear_interpolation')

    naming_convention = Enum('line', 'grid_serial', 'grid_concatenate')
    grid_type = Enum('depth', 'subdural')
    label_auto_action = Action(name='Automatic labeling',
        action='do_label_automatically')
    label_manually_action = Action(name='Label with current geometry',
        action='do_label_manually')

    #have grid type reflect the type of grid input and also change the
    #default naming convention

    #make these naming conventions apply

    #then work on documentation

    rotate_grid_left_action = Action(name='Rotate left 90 deg',
        action='do_rotate_left')
    rotate_grid_right_action = Action(name='Rotate right 90 deg',
        action='do_rotate_right')
    rotate_grid_180_action = Action(name='Rotate 180 deg',
        action='do_rotate_180')
    reflect_grid_action = Action(name='Reflect',
        action='do_reflect')

    name_stem = Str
    c1, c2, c3 = 3*(Instance(Electrode),)

    parcellation = Str
    error_radius = Float(4)
    find_rois_action = Action(name='Estimate single ROI contacts',
        action='do_rois')
    find_all_rois_action = Action(name='Estimate all ROI contacts',
        action='do_all_rois')

    manual_reposition_action = Action(name='Manually modify electrode '
        'position', action='do_manual_reposition')

    img_dpi = Float(125.)
    img_size = List(Float)
    save_coronal_slice_action = Action(name='Save coronal slice',
        action='do_coronal_slice')

    labeling_delta = Float(0.25)
    labeling_epsilon = Int(25)
    labeling_rho = Int(40)

    include_underscore = Bool(True)

    def _img_size_default(self):
        return [450., 450.]

    #electrode_factory = Method
    def electrode_factory(self):
        return Electrode(special_name='Electrode for linear interpolation',
            grid_name=self.cur_grid,
            is_interpolation=True)

    def dynamic_view(self):
        return View(
            Item('electrodes',
                editor=TableEditor( columns = 
                    [ObjectColumn(label='electrode',
                                  editor=TextEditor(),
                                  style='readonly',
                                  editable=False,
                                  name='default_repr'),

                     ObjectColumn(label='corner',
                                  editor=CheckListEditor(
                                    values=['','corner 1','corner 2',
                                        'corner 3']),
                                  style='simple',
                                  name='corner'),

                     ObjectColumn(label='geometry',
                                  editor=CSVListEditor(),
                                  name='geom_coords'),
                                  
                     ObjectColumn(label='channel name',
                                  editor=TextEditor(),
                                  name='name'),

                     ObjectColumn(label='ROIs',
                                  editor=ListStrEditor(),
                                  editable=False, 
                                  name='roi_list'),

                     ObjectColumn(label='RAS',
                                  editor=TextEditor(),
                                  style='readonly',
                                  editable=False,
                                  name='ras_repr'),

                     ObjectColumn(label='iso',
                                  editor=TextEditor(),
                                  style='readonly',
                                  editable=False,
                                  name='iso_repr'),

                     ObjectColumn(label='snap',
                                  editor=TextEditor(),
                                  style='readonly',
                                  editable=False,
                                  name='postsnap_repr'),
                     ],
                    selected='cur_sel',
                    deletable=True,
                    row_factory=self.electrode_factory,
                    ),
                show_label=False, height=350, width=700),

            HGroup(
                VGroup( 
                    Label( 'Automatic labeling parameters' ),
                    HGroup(
                        Item( 'name_stem' ),
                        Item( 'naming_convention' ),
                        Item( 'include_underscore' ),
                        Item( 'grid_type' ),
                    ),
                    HGroup(
                        Item( 'labeling_delta' ),
                        Item( 'labeling_rho' ),
                        Item( 'labeling_epsilon' ),
                    ),
                ),
                #VGroup(
                #    Label( 'ROI identification parameters' ),
                #    Item('parcellation'),
                #    Item('error_radius'),
                #),
                #VGroup(
                #    Label('Image parameters' ),
                #    Item('img_dpi', label='dpi'),
                #    Item('img_size', label='size', editor=CSVListEditor()),
                #),
            ),

            resizable=True, kind='panel', title='modify electrodes',
            #buttons=[OKButton, swap_action, label_auto_action,
            #    interpolate_action, save_montage_action, find_rois_action]) 

            buttons = [self.label_auto_action, self.swap_action, OKButton],
            menubar = MenuBar(
                Menu( self.label_auto_action,
                      self.label_manually_action,
                      self.rotate_grid_left_action,
                      self.rotate_grid_right_action,
                      self.rotate_grid_180_action,
                      self.reflect_grid_action,
                      name='Labeling',
                ),
                Menu( self.add_blank_action,
                      self.interpolate_action, 
                      self.find_rois_action, 
                      self.find_all_rois_action,
                      self.manual_reposition_action,
                      name='Operations',
                ),
                Menu( self.save_montage_action, 
                      self.save_csv_action,
                      self.save_coronal_slice_action,
                      name='Save Output',
                ),
            )
        )

    def edit_traits(self):
        super(ElectrodeWindow, self).edit_traits(view=self.dynamic_view())

    @on_trait_change('cur_sel')
    def selection_callback(self):
        if self.cur_sel is None:
            return

        #import pdb
        #pdb.set_trace()

        if self.previous_sel is not None:
            self.model._new_glyph_color = self.previous_color
            self.model._single_glyph_to_recolor = self.previous_sel.asiso()
            self.model._update_single_glyph_event = True


        # if this electrode has just been created and not interpolated yet
        # it has no change of being in the image yet so just pass
        if self.cur_sel.special_name == 'Electrode for linear interpolation':
            self.previous_sel = None 
            self.distinct_prev_sel = None
            return

        if self.distinct_prev_sel != self.previous_sel:
            self.distinct_prev_sel = self.previous_sel

        self.previous_sel = self.cur_sel
        self.previous_color = list(self.model._colors.keys()).index(self.cur_grid)

        selection_color = (list(self.model._colors.keys()).index('selection'))

        self.model._new_glyph_color = selection_color
        self.model._single_glyph_to_recolor = self.cur_sel.asiso()
        self.model._update_single_glyph_event = True

    #whenever the window closes, it no longer has a valid cur_sel to listen to
    def closed(self, is_ok, info):
        self.cur_sel = None 
        del self.model.ews[self.cur_grid]
        if self.previous_sel is not None:
            self.model._new_glyph_color = self.previous_color
            self.model._single_glyph_to_recolor = self.previous_sel.asiso()
            self.model._update_single_glyph_event = True

    @on_trait_change('grid_type')
    def change_grid_type(self):
        self.model._grid_types[self.cur_grid] = self.grid_type

    def do_add_blank(self, info):
        e = self.electrode_factory()
        e.grid_name = self.cur_grid
        self.electrodes.append(e)

    def do_swap(self, info):
        #if not len(self.selected_ixes) == 2:
        #    return
        if self.distinct_prev_sel == self.cur_sel:
            return
        elif None in (self.distinct_prev_sel, self.cur_sel):
            return

        #i,j = self.selected_ixes
        #e1 = self.electrodes[i]
        #e2 = self.electrodes[j]
        e1 = self.cur_sel
        e2 = self.distinct_prev_sel

        geom_swap = e1.geom_coords
        name_swap = e1.name

        e1.geom_coords = e2.geom_coords
        e1.name = e2.name

        e2.geom_coords = geom_swap
        e2.name = name_swap

    def do_label_automatically(self, info):
        #figure out c1, c2, c3
#        c1,c2,c3 = 3*(None,)
#        for e in self.electrodes:
#            if len(e.corner) == 0:
#                continue
#            elif len(e.corner) > 1:
#                error_dialog('Wrong corners specified, check again')
#                return
#    
#            elif 'corner 1' in e.corner:
#                c1 = e
#            elif 'corner 2' in e.corner:
#                c2 = e
#            elif 'corner 3' in e.corner:
#                c3 = e
#
#        if c1 is None or c2 is None or c3 is None:
#            error_dialog('Not all corners were specified')
#            return
    
        cur_geom = self.model._grid_geom[self.cur_grid]
        if cur_geom=='user-defined' and self.naming_convention != 'line':
            from .color_utils import mayavi2traits_color
            from .name_holder import GeometryNameHolder, GeomGetterWindow
            nameholder = GeometryNameHolder(
                geometry=cur_geom,
                color=mayavi2traits_color(
                    self.model._colors[self.cur_grid]))
            geomgetterwindow = GeomGetterWindow(holder=nameholder)

            if geomgetterwindow.edit_traits().result:
                cur_geom = geomgetterwindow.geometry
            else:
                error_dialog("User did not specify any geometry")
                return

        from . import pipeline as pipe
        if self.naming_convention == 'line':
            pipe.fit_grid_to_line(self.electrodes, 
                #c1.asiso(), c2.asiso(), c3.asiso(), cur_geom, 
                delta=self.labeling_delta,
                rho_loose=self.labeling_rho+10,
                epsilon=self.labeling_epsilon)

        else:
            pipe.fit_grid_by_fixed_points(self.electrodes, cur_geom,
                delta=self.labeling_delta, 
                rho=self.labeling_rho, 
                rho_strict=self.labeling_rho-10, 
                rho_loose=self.labeling_rho+15,
                epsilon=self.labeling_epsilon, 
                mindist=0, maxdist=36)

        self.naming_following_labeling(cur_geom=cur_geom)

    def do_labeling_manually(self):
        cur_geom = self.model._grid_geom[self.cur_grid]
        self.naming_following_labeling(cur_geom=cur_geom)

    def naming_following_labeling(self, cur_geom=None):
        xmax, ymax = (0, 0)

        for elec in self.electrodes:
            x, y = elec.geom_coords

            xmax = np.max((x, xmax))
            ymax = np.max((y, ymax))

        for elec in self.electrodes:
            x, y = elec.geom_coords

            if self.naming_convention == 'grid_serial':
                if xmax > ymax:
                    index = x * np.min(cur_geom) + y + 1
                else:
                    index = x * np.max(cur_geom) + y + 1
            elif self.naming_convention == 'grid_concatenate':
                index = '{0}{1}'.format(x + 1, y + 1)

            elif self.naming_convention == 'line':
                index = y+1

            elec.name = '{0}{1}{2}'.format(self.name_stem, 
                '_' if self.include_underscore else '', 
                index)

    def do_linear_interpolation(self, info):
        #TODO does not feed back coordinates to model, 
        #which is important for snapping

        if self.cur_sel is None:
            return
        elif self.cur_sel.special_name == '':
            return
        
        if len(self.cur_sel.geom_coords) == 0:
            error_dialog("Specify geom_coords before linear interpolation")
            return

        x,y = self.cur_sel.geom_coords

        x_low = self._find_closest_neighbor(self.cur_sel, 'x', '-')
        x_hi = self._find_closest_neighbor(self.cur_sel, 'x', '+')
        y_low = self._find_closest_neighbor(self.cur_sel, 'y', '-')
        y_hi = self._find_closest_neighbor(self.cur_sel, 'y', '+')

        loc = None

        #handle simplest case of electrode directly in between others
        if x_low is not None and x_hi is not None:
            xl = x_low.geom_coords[0]
            xh = x_hi.geom_coords[0]
            ratio = (x - xl) / (xh - xl)
        
            loc = np.array(x_low.iso_coords) + (np.array(x_hi.iso_coords)-
                np.array(x_low.iso_coords))*ratio

        elif y_low is not None and y_hi is not None:
            yl = y_low.geom_coords[1]
            yh = y_hi.geom_coords[1]
            ratio = (y - yl) / (yh - yl)
        
            loc = np.array(y_low.iso_coords) + (np.array(y_hi.iso_coords)-
                np.array(y_low.iso_coords))*ratio

        #handle poorer case of electrode on end of line
        if x_low is not None and loc is None:
            x_lower = self._find_closest_neighbor(x_low, 'x', '-')
            xl = x_low.geom_coords[0]
            xll = x_lower.geom_coords[0]
            if xl == xll+1:
                loc = 2*np.array(x_low.iso_coords) - np.array(
                    x_lower.iso_coords)

        if x_hi is not None and loc is None:
            x_higher = self._find_closest_neighbor(x_hi, 'x', '+')
            xh = x_hi.geom_coords[0]
            xhh = x_higher.geom_coords[0]
            if xh == xhh-1:
                loc = 2*np.array(x_hi.iso_coords) - np.array(
                    x_higher.iso_coords)

        #import pdb
        #pdb.set_trace()

        if y_low is not None and loc is None:
            y_lower = self._find_closest_neighbor(y_low, 'y', '-')
            yl = y_low.geom_coords[1]
            yll = y_lower.geom_coords[1]
            if yl == yll+1:
                loc = 2*np.array(y_low.iso_coords) - np.array(
                    y_lower.iso_coords)
        
        if y_hi is not None and loc is None:
            y_higher = self._find_closest_neighbor(y_hi, 'y', '+')
            yh = y_hi.geom_coords[1]
            yhh = y_higher.geom_coords[1]
            if yh == yhh-1:
                loc = 2*np.array(y_hi.iso_coords) - np.array(
                    y_higher.iso_coords)
    
        if loc is not None:
            self.cur_sel.iso_coords = tuple(loc)
            self.cur_sel.special_name = 'Linearly interpolated electrode'
        else:
            error_dialog('No line for simple linear interpolation\n'
                'Better algorithm needed')

        # translate the electrode into RAS space
        from . import pipeline as pipe

        pipe.linearly_transform_electrodes_to_isotropic_coordinate_space(
            [self.cur_sel], self.model.ct_scan,
            isotropization_strategy = self.model.isotropize,
            isotropization_direction_off = 'copy_to_ct',
            isotropization_direction_on = 'deisotropize',
            iso_vector_override = self.model.isotropization_override)
        
        aff = self.model.acquire_affine()

        pipe.translate_electrodes_to_surface_space( [self.cur_sel], aff,
            subjects_dir=self.model.subjects_dir, subject=self.model.subject)

        # add this electrode to the grid model so that it can be visualized
        self.model.add_electrode_to_grid(self.cur_sel, self.cur_grid)

    def _find_closest_neighbor(self, cur_elec, axis, direction): 
        x,y = cur_elec.geom_coords

        if direction=='+':
            new_ix = np.inf
        else:
            new_ix = -np.inf
        new_e = None

        for e in self.electrodes:
            if len(e.geom_coords) == 0:
                continue

            ex,ey = e.geom_coords
            
            if axis=='x' and direction=='+':
                if ex < new_ix and ex > x and ey == y:
                    new_e = e
                    new_ix = ex
            if axis=='x' and direction=='-':
                if ex > new_ix and ex < x and ey == y:
                    new_e = e
                    new_ix = ex
            if axis=='y' and direction=='+':
                if ey < new_ix and ey > y and ex == x:
                    new_e = e
                    new_ix = ey
            if axis=='y' and direction=='-':
                if ey > new_ix and ey < y and ex == x:
                    new_e = e
                    new_ix = ey

        return new_e

    def do_rotate_left(self, info):
        electrodes = self.model.get_electrodes_from_grid(
            target=self.cur_grid,
            electrodes=self.electrodes)

        # brute force check geom coords from actual data
        # getting geom coords from the model wouldn't ensure the order
        geom_x, geom_y = -1, -1
        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            if curx > geom_x:
                geom_x = curx
            if cury > geom_y:
                geom_y = cury
            
        # now change the coordinates
        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            newx = cury
            newy = geom_x - curx
            elec.geom_coords = [newx, newy]
   
    def do_rotate_right(self, info):
        electrodes = self.model.get_electrodes_from_grid(
            target=self.cur_grid,
            electrodes=self.electrodes)

        # brute force check geom coords from actual data
        # getting geom coords from the model wouldn't ensure the order
        geom_x, geom_y = -1, -1
        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            if curx > geom_x:
                geom_x = curx
            if cury > geom_y:
                geom_y = cury
            
        # now change the coordinates
        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            newx = geom_y - cury
            newy = curx
            elec.geom_coords = [newx, newy]

    def do_rotate_180(self, info):
        electrodes = self.model.get_electrodes_from_grid(
            target=self.cur_grid,
            electrodes=self.electrodes)

        # brute force check geom coords from actual data
        # getting geom coords from the model wouldn't ensure the order
        geom_x, geom_y = -1, -1
        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            if curx > geom_x:
                geom_x = curx
            if cury > geom_y:
                geom_y = cury
            
        # now change the coordinates
        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            newx = geom_x - curx
            newy = geom_y - cury
            elec.geom_coords = [newx, newy]

    def do_reflect(self, info):
        electrodes = self.model.get_electrodes_from_grid(
            target=self.cur_grid,
            electrodes=self.electrodes)

        for elec in electrodes:
            if len(elec.geom_coords) == 0:
                continue
            curx, cury = elec.geom_coords
            newx = cury
            newy = curx
            elec.geom_coords = [newx, newy]

    def do_montage(self, info):
        electrodes = self.model.get_electrodes_from_grid(
            target=self.cur_grid,
            electrodes=self.electrodes)

        if electrodes is None:
            return

        from .electrode_group import save_coordinates
        save_coordinates( electrodes, self.model._grid_types,
            snapping_completed=self.model._snapping_completed,
            file_type='montage')

    def do_csv(self, info):
        electrodes = self.model.get_electrodes_from_grid(
            target=self.cur_grid,
            electrodes=self.electrodes)

        if electrodes is None:
            return

        from .electrode_group import save_coordinates
        save_coordinates( electrodes, self.model._grid_types,
            snapping_completed=self.model._snapping_completed,
            file_type='csv')

    def do_rois(self, info):
        from .electrode_group import get_nearby_rois_elec
        get_nearby_rois_elec( self.cur_sel,
            parcellation=self.model.roi_parcellation,
            error_radius=self.model.roi_error_radius,
            subjects_dir=self.model.subjects_dir, 
            subject=self.model.subject )

    def do_all_rois(self, info):
        from .electrode_group import get_nearby_rois_grid
        get_nearby_rois_grid( self.electrodes,
            parcellation=self.model.roi_parcellation,
            error_radius=self.model.roi_error_radius,
            subjects_dir=self.model.subjects_dir, 
            subject=self.model.subject )

    def do_coronal_slice(self, info):
        savefile = ask_user_for_savefile('save png file with slice image')

        from .electrode_group import coronal_slice_grid
        coronal_slice_grid(self.electrodes, savefile=savefile,
            subjects_dir=self.model.subjects_dir, subject=self.model.subject,
            dpi=self.model.coronal_dpi, 
            size=tuple(self.model.coronal_size),
            title=self.name_stem)

    def do_manual_reposition(self, info):
        if self.cur_sel is None:
            return

        pd = self.model.construct_panel2d()
        #import panel2d

        x,y,z = self.cur_sel.asct()
        pd.move_cursor(x,y,z)
        pd.drop_pin(x,y,z, color='cyan', name='electrode', image_name='ct')

        rx,ry,rz = self.cur_sel.asras()
        pd.drop_pin(rx,ry,rz, color='cyan', name='electrode', image_name='t1',
            ras_coords=True)

        pd.edit_traits(kind='livemodal')

    @on_trait_change('model:panel2d:move_electrode_internally_event')
    def _internally_effect_electrode_reposition(self):
        if self.cur_sel is None:
            error_dialog("No electrode specified to move")
            return

        pd = self.model.panel2d
        image_name = pd.currently_showing.name
        px,py,pz,_ = pd.pins[image_name][pd.current_pin]

        if image_name=='t1':
            px,py,pz = pd.map_cursor((px,py,pz), pd.images['t1'][2])
            
        self.model.move_electrode( self.cur_sel, (px,py,pz),
            in_ras=(image_name=='t1') )

    @on_trait_change('model:panel2d:move_electrode_postprocessing_event')
    def _postprocessing_effect_electrode_reposition(self):
        if self.cur_sel is None:
            error_dialog("No electrode specified to move")
            return

        pd = self.model.panel2d
        image_name = pd.currently_showing.name
        px,py,pz,_ = pd.pins[image_name][pd.current_pin]

        if image_name=='t1':
            px,py,pz = pd.map_cursor((px,py,pz), pd.images['t1'][2])
            
        self.model.move_electrode( self.cur_sel, (px,py,pz),
            in_ras=(image_name=='t1'), as_postprocessing=True )



import os
import setuptools


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setuptools.setup(
    name="ielu",
    version="0.3.0",
    maintainer="Roan LaPlante",
    maintainer_email="rlaplant@nmr.mgh.harvard.edu",
    description=("Interactive Electrode localization Utility"),
    license="Visuddhimagga Sutta; GPLv3+",
    datafiles=[('', ['readme.md', 'LICENSE'])],
    long_description=read('readme.md'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: X11 Applications",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Information Analysis",
    ],
    url="https://github.com/aestrivex/ielu",
    platforms=['any'],
    packages=['ielu'],
    package_data={'ielu':['make_dural_surface.csh']},
    scripts=['bin/ielu', 'ielu/mkoutersurf.py'],
    install_requires=["numpy", "scipy", "pymcubes", "pysurfer", "nibabel", "mne"]
)
